from qtpy.QtWidgets import QFileDialog
import json
import logging

from ..utilities.get import Get
from ..session import SessionKeys
from .utilities import create_bin_tab_output_file_name
from ..utilities.status_message_config import StatusMessageStatus, show_status_message


class ExportBinTable:
    """
    this class will export as a json file the dictionary of bins (file_index, tof and lambda)
    """

    def __init__(self, parent=None):
        self.parent = parent
        self.logger = logging.getLogger("maverick")

    def run(self):

        working_dir = self.parent.session[SessionKeys.top_folder]
        _folder = str(QFileDialog.getExistingDirectory(caption="Select Folder to ExportImages the Images",
                                                       directory=working_dir,
                                                       options=QFileDialog.ShowDirsOnly))

        if _folder == "":
            self.logger.info("User cancel export bin table!")
            return

        o_get = Get(parent=self.parent)
        current_bin_name_activated = o_get.current_bins_name_activated()
        self.logger.info(f"{current_bin_name_activated} bins table will be exported to {_folder}!")

        current_bins = o_get.current_bins_activated()
        output_file_name = create_bin_tab_output_file_name(folder=_folder,
                                                           bin_name=current_bin_name_activated)

        with open(output_file_name, 'w') as json_file:
            json.dump(current_bins, json_file)

        show_status_message(parent=self.parent,
                            message=f"Exported bin table as {output_file_name} ... Done!",
                            status=StatusMessageStatus.ready,
                            duration_s=10)
        self.logger.info(f"Bin table {current_bin_name_activated} has been exported as {output_file_name}")



from ..utilities import BinMode, CombineAlgorithm


class SessionKeys:
    """list of all sessions keys, to easily retrieve them"""

    top_folder = "top_folder"
    list_working_folders = "list_working_folders"
    list_working_folders_status = "list_working_folders_status"
    log_buffer_size = "log_buffer_size"
    version = "version"
    distance_source_detector = 'distance_source_detector'
    detector_offset = 'detector_offset'
    combine_algorithm = 'combine_algorithm'
    combine_roi = 'combine_roi'
    sample_position = 'sample_position'
    bin_mode = 'bin_mode'
    bin_algorithm = 'bin_algorithm'


session = {SessionKeys.top_folder: None,  # the base folder to start looking at images folder to combine
           SessionKeys.list_working_folders: None,  # list of working folders
           SessionKeys.list_working_folders_status: None,  # list of working folders status [True, True, False..]
           SessionKeys.log_buffer_size: 500,   # max size of the log file
           SessionKeys.version: "0.0.1",   # version of that config
           SessionKeys.distance_source_detector: 19.855,
           SessionKeys.detector_offset: 9600,
           SessionKeys.combine_algorithm: CombineAlgorithm.mean,
           SessionKeys.combine_roi: [50, 50, 200, 200],  # [x0,y0,width,height],
           SessionKeys.sample_position: 0,    # in the combine tab
           SessionKeys.bin_mode: BinMode.auto,   # 'auto' or 'manual',
           SessionKeys.bin_algorithm: CombineAlgorithm.mean,  # 'mean' or 'median'
           }



import logging
import pyqtgraph as pg

from ..utilities import TimeSpectraKeys
from .plot import Plot
from ..utilities.get import Get
from . import TO_MICROS_UNITS, TO_ANGSTROMS_UNITS
from ..utilities.table_handler import TableHandler
from ..utilities.math_tools import get_value_of_closest_match
from ..utilities.string import format_str

FILE_INDEX_BIN_MARGIN = 0.5
UNSELECTED_BIN = (0, 0, 200, 50)
SELECTED_BIN = (0, 200, 0, 50)


class ManualEventHandler:

    def __init__(self, parent=None):
        self.parent = parent
        self.logger = logging.getLogger('maverick')

        self.tof_bin_margin = (self.parent.time_spectra[TimeSpectraKeys.tof_array][1] -
                               self.parent.time_spectra[TimeSpectraKeys.tof_array][0]) / 2.

        self.lambda_bin_margin = (self.parent.time_spectra[TimeSpectraKeys.lambda_array][1] -
                                  self.parent.time_spectra[TimeSpectraKeys.lambda_array][0]) / 2

    def refresh_manual_tab(self):
        """refresh the right plot with profile + bin selected when the manual tab is selected"""
        o_plot = Plot(parent=self.parent)
        o_plot.refresh_profile_plot()

    def add_bin(self):

        o_get = Get(parent=self.parent)
        time_spectra_x_axis_name = o_get.bin_x_axis_selected()
        x_axis = self.parent.time_spectra[time_spectra_x_axis_name]

        o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)
        last_row = o_table.row_count()

        if time_spectra_x_axis_name == TimeSpectraKeys.file_index_array:
            default_bin = [x_axis[0] - FILE_INDEX_BIN_MARGIN,
                           x_axis[0] + FILE_INDEX_BIN_MARGIN]
        elif time_spectra_x_axis_name == TimeSpectraKeys.tof_array:
            default_bin = [x_axis[0] - self.tof_bin_margin,
                           x_axis[0] + self.tof_bin_margin]
            default_bin = [_value * TO_MICROS_UNITS for _value in default_bin]
        else:
            default_bin = [x_axis[0] - self.lambda_bin_margin,
                           x_axis[0] + self.lambda_bin_margin]
            default_bin = [_value * TO_ANGSTROMS_UNITS for _value in default_bin]

        manual_bins = self.parent.manual_bins
        if manual_bins[TimeSpectraKeys.file_index_array] is None:
            manual_bins[TimeSpectraKeys.file_index_array] = [[x_axis[0]]]
            manual_bins[TimeSpectraKeys.tof_array] = [[self.parent.time_spectra[TimeSpectraKeys.tof_array][0]]]
            manual_bins[TimeSpectraKeys.lambda_array] = [[self.parent.time_spectra[TimeSpectraKeys.lambda_array][0]]]
        else:
            manual_bins[TimeSpectraKeys.file_index_array].append([x_axis[0]])
            manual_bins[TimeSpectraKeys.tof_array].append([self.parent.time_spectra[TimeSpectraKeys.tof_array][0]])
            manual_bins[TimeSpectraKeys.lambda_array].append([self.parent.time_spectra[TimeSpectraKeys.lambda_array][0]])
        self.parent.manual_bins = manual_bins

        item = pg.LinearRegionItem(values=default_bin,
                                   orientation='vertical',
                                   brush=SELECTED_BIN,
                                   movable=True,
                                   bounds=None)
        item.setZValue(-10)
        item.sigRegionChangeFinished.connect(self.parent.bin_manual_region_changed)

        self.parent.bin_profile_view.addItem(item)
        # dict_of_bins_item[last_row] = item
        # self.parent.dict_of_bins_item = dict_of_bins_item
        self.parent.list_of_manual_bins_item.append(item)

        # add new entry in table
        o_table.insert_empty_row(last_row)

        o_table.insert_item(row=last_row,
                            column=0,
                            value=f"{last_row}",
                            editable=False)

        _file_index = self.parent.time_spectra[TimeSpectraKeys.file_index_array][0]
        o_table.insert_item(row=last_row,
                            column=1,
                            value=_file_index,
                            editable=False)

        _tof = self.parent.time_spectra[TimeSpectraKeys.tof_array][0] * TO_MICROS_UNITS
        o_table.insert_item(row=last_row,
                            column=2,
                            value=_tof,
                            format_str="{:.2f}",
                            editable=False)

        _lambda = self.parent.time_spectra[TimeSpectraKeys.lambda_array][0] * TO_ANGSTROMS_UNITS
        o_table.insert_item(row=last_row,
                            column=3,
                            value=_lambda,
                            format_str="{:.3f}",
                            editable=False)

    def clear_all_items(self):
        list_of_manually_bins_item = self.parent.list_of_manual_bins_item
        for _item in list_of_manually_bins_item:
            self.parent.bin_profile_view.removeItem(_item)

        self.parent.list_of_manual_bins_item = None

    def display_all_items(self):
        list_of_manually_bins_item = self.parent.list_of_manual_bins_item
        for _item in list_of_manually_bins_item:
            self.parent.bin_profile_view.addItem(_item)

    def populate_table_with_this_table(self, table=None):
        o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)
        o_table.remove_all_rows()

        file_index_array = table[TimeSpectraKeys.file_index_array]
        tof_array = table[TimeSpectraKeys.tof_array]
        lambda_array = table[TimeSpectraKeys.lambda_array]

        self.parent.list_of_manual_bins_item = []

        if file_index_array is None:
            return

        _row = 0
        for _index, _bin in enumerate(file_index_array):

            if _bin == []:
                continue

            o_table.insert_empty_row(_row)

            o_table.insert_item(row=_row,
                                column=0,
                                value=f"{_row}",
                                editable=False)

            _file_index = _bin
            _file_index_formatted = format_str(_file_index,
                                               format_str="{:d}",
                                               factor=1,
                                               data_type=TimeSpectraKeys.file_index_array)
            o_table.insert_item(row=_row,
                                column=1,
                                format_str=_file_index_formatted,
                                editable=False)

            _tof = tof_array[_index]
            _tof_formatted = format_str(_tof,
                                        format_str="{:.2f}",
                                        factor=TO_MICROS_UNITS,
                                        data_type=TimeSpectraKeys.tof_array)

            o_table.insert_item(row=_row,
                                column=2,
                                format_str=_tof_formatted,
                                editable=False)

            _lambda = lambda_array[_index]
            _lambda_formatted = format_str(_lambda,
                                           format_str="{:.3f}",
                                           factor=TO_ANGSTROMS_UNITS,
                                           data_type=TimeSpectraKeys.lambda_array)
            o_table.insert_item(row=_row,
                                column=3,
                                format_str=_lambda_formatted,
                                editable=False)

            item = self.add_bin_in_plot(row=_row,
                                        file_index_bin=_bin,
                                        tof_bin=_tof,
                                        lambda_bin=_lambda)

            self.parent.list_of_manual_bins_item.append(item)

            _row += 1

        o_table.select_rows([0])

    def populate_table_with_auto_mode(self):
        o_get = Get(parent=self.parent)
        bins = o_get.auto_bins_currently_activated()
        self.parent.manual_bins = bins
        self.populate_table_with_this_table(table=bins)

    def add_bin_in_plot(self, row=0, file_index_bin=None, tof_bin=None, lambda_bin=None):
        o_get = Get(parent=self.parent)
        current_x_axis = o_get.bin_x_axis_selected()
        if current_x_axis == TimeSpectraKeys.file_index_array:
            bin = file_index_bin
            bin_size = [bin[0] - FILE_INDEX_BIN_MARGIN,
                        bin[-1] + FILE_INDEX_BIN_MARGIN]
        elif current_x_axis == TimeSpectraKeys.tof_array:
            bin = tof_bin
            bin_size = [bin[0] - self.tof_bin_margin,
                        bin[-1] + self.tof_bin_margin]
        elif current_x_axis == TimeSpectraKeys.lambda_array:
            bin = lambda_bin
            bin_size = [bin[0] - self.lambda_bin_margin,
                        bin[1] + self.lambda_bin_margin]
        else:
            raise NotImplementedError("x_axis not implemented!")

        if row == 0:
            brush_selection = SELECTED_BIN
        else:
            brush_selection = UNSELECTED_BIN

        item = pg.LinearRegionItem(values=bin_size,
                                   orientation='vertical',
                                   brush=brush_selection,
                                   movable=True,
                                   bounds=None)
        item.setZValue(-10)
        item.sigRegionChangeFinished.connect(self.parent.bin_manual_region_changed)
        self.parent.bin_profile_view.addItem(item)

        return item

    # def manual_table_right_click(self):
    #     o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)
    #     last_row = o_table.row_count()
    #
    #     menu = QMenu(self.parent)
    #
    #     row_selected = o_table.get_row_selected()
    #     remove_bin = -1
    #     # clean_sort = None
    #     load_table = menu.addAction("Import table ...")
    #
    #     if last_row > 0:
    #         menu.addSeparator()
    #         remove_bin = menu.addAction("Remove selected bin")
    #         # clean_sort = menu.addAction("Sort and remove duplicates")
    #
    #     action = menu.exec_(QtGui.QCursor.pos())
    #     if action == remove_bin:
    #         self.remove_selected_bin()
    #         self.parent.update_statistics()
    #     elif action == load_table:
    #         self.load_manual_bin_table()
    #     # elif action == clean_sort:
    #     #     self.sort_and_remove_duplicates()
    #
    #     else:
    #         pass
    #
    # def load_manual_bin_table(self):
    #     o_load = LoadBinTable(parent=self.parent)
    #     o_load.run()

    # def remove_selected_bin(self):
    #     """
    #     remove from the manual table the bin selected
    #     """
    #     o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)
    #     row_selected = o_table.get_row_selected()
    #     item_to_remove = self.parent.list_of_manual_bins_item[row_selected]
    #     self.parent.bin_profile_view.removeItem(item_to_remove)
    #     self.parent.list_of_manual_bins_item.pop(row_selected)
    #     o_table.remove_row(row=row_selected)
    #     self.logger.info(f"User manually removed row: {row_selected}")
    #
    # def sort_and_remove_duplicates(self):
    #     print("sort and remove duplicates")

    def bin_manually_moved(self, item_id=None):
        self.bin_manually_moving(item_id=item_id)

        # 1. using region selected threshold, and the current axis, find the snapping left and right indexes
        #    and save them into a manual_snapping_indexes_bins = {0: [0, 3], 1: [1, 10], ..}
        self.record_snapping_indexes_bin()

        # 2. reposition the clean bins into the plot
        self.update_items_displayed()

        # 3. using those indexes create the ranges for each bins and for each time axis and save those in
        #    self.parent.manual_bins['file_index_array': {0: [0, 1, 2, 3], 1: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], ...},...}
        self.create_all_ranges()

        # 4. update table
        self.update_table()

    def bin_manually_moving(self, item_id=None):
        o_get = Get(parent=self.parent)
        working_row = o_get.manual_working_row(working_item_id=item_id)
        self.select_working_row(working_row=working_row)

    def select_working_row(self, working_row=0):
        o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)
        o_table.select_rows(list_of_rows=[working_row])

    def update_table(self):

        o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)

        file_index_array = self.parent.manual_bins[TimeSpectraKeys.file_index_array]
        tof_array = self.parent.manual_bins[TimeSpectraKeys.tof_array]
        lambda_array = self.parent.manual_bins[TimeSpectraKeys.lambda_array]

        for _row in file_index_array.keys():
            list_runs = file_index_array[_row]
            list_runs_formatted = format_str(list_runs,
                                             format_str="{:d}",
                                             factor=1,
                                             data_type=TimeSpectraKeys.file_index_array)
            o_table.set_item_with_str(row=_row, column=1, cell_str=list_runs_formatted)

            list_tof = tof_array[_row]
            list_tof_formatted = format_str(list_tof,
                                            format_str="{:.2f}",
                                            factor=TO_MICROS_UNITS,
                                            data_type=TimeSpectraKeys.tof_array)
            o_table.set_item_with_str(row=_row, column=2, cell_str=list_tof_formatted)

            list_lambda = lambda_array[_row]
            list_lambda_formatted = format_str(list_lambda,
                                               format_str="{:.3f}",
                                               factor=TO_ANGSTROMS_UNITS,
                                               data_type=TimeSpectraKeys.lambda_array)
            o_table.set_item_with_str(row=_row, column=3, cell_str=list_lambda_formatted)

    def create_all_ranges(self):
        manual_snapping_indexes_bins = self.parent.manual_snapping_indexes_bins

        file_index_array = {}
        tof_array = {}
        lambda_array = {}

        for _bin in manual_snapping_indexes_bins.keys():
            left_index, right_index = manual_snapping_indexes_bins[_bin]

            # tof_array
            bins_file_index_array = self.parent.time_spectra[TimeSpectraKeys.file_index_array]
            bins_file_index_range = bins_file_index_array[left_index: right_index + 1]
            file_index_array[_bin] = bins_file_index_range

            # tof_array
            bins_tof_array = self.parent.time_spectra[TimeSpectraKeys.tof_array]
            bins_tof_range = bins_tof_array[left_index: right_index + 1]
            tof_array[_bin] = bins_tof_range

            # lambda_array
            bins_lambda_array = self.parent.time_spectra[TimeSpectraKeys.lambda_array]
            bins_lambda_range = bins_lambda_array[left_index: right_index + 1]
            lambda_array[_bin] = bins_lambda_range

        self.parent.manual_bins[TimeSpectraKeys.file_index_array] = file_index_array
        self.parent.manual_bins[TimeSpectraKeys.tof_array] = tof_array
        self.parent.manual_bins[TimeSpectraKeys.lambda_array] = lambda_array

    def update_items_displayed(self):
        """
        this will remove the old item and put the new one with the edges snap to the x-axis
        """
        o_get = Get(parent=self.parent)
        x_axis_type_selected = o_get.bin_x_axis_selected()
        x_axis = self.parent.time_spectra[x_axis_type_selected]

        manual_snapping_indexes_bins = self.parent.manual_snapping_indexes_bins

        margin = self.margin()

        list_of_manual_bins_item = []
        for _row in manual_snapping_indexes_bins.keys():
            left_value_checked, right_value_checked = manual_snapping_indexes_bins[_row]
            left_value_checked = x_axis[left_value_checked] - margin
            right_value_checked = x_axis[right_value_checked] + margin

            _item = self.parent.list_of_manual_bins_item[_row]
            self.parent.bin_profile_view.removeItem(_item)

            item = pg.LinearRegionItem(values=[left_value_checked, right_value_checked],
                                       orientation='vertical',
                                       brush=SELECTED_BIN,
                                       movable=True,
                                       bounds=None)
            item.setZValue(-10)
            item.sigRegionChangeFinished.connect(self.parent.bin_manual_region_changed)
            item.sigRegionChanged.connect(self.parent.bin_manual_region_changing)
            self.parent.bin_profile_view.addItem(item)
            list_of_manual_bins_item.append(item)

        self.parent.list_of_manual_bins_item = list_of_manual_bins_item

    def record_bin_ranges(self):
        """
        record all the bins ranges in all the x-axis values
        """
        pass

    def record_snapping_indexes_bin(self):
        """
        This will check each bin from the manual table and move, if necessary, any of the edges
        to snap to the closet x axis values
        """
        manual_snapping_indexes_bins = {}
        for _row, _item in enumerate(self.parent.list_of_manual_bins_item):
            [left, right] = _item.getRegion()

            # bring left and right to closest correct values
            left_value_checked, right_value_checked = self.checked_range(left=left, right=right)
            manual_snapping_indexes_bins[_row] = [left_value_checked, right_value_checked]

        self.parent.manual_snapping_indexes_bins = manual_snapping_indexes_bins

    def margin(self, axis_type=TimeSpectraKeys.file_index_array):
        if axis_type == TimeSpectraKeys.file_index_array:
            return FILE_INDEX_BIN_MARGIN
        elif axis_type == TimeSpectraKeys.tof_array:
            return self.tof_bin_margin
        elif axis_type == TimeSpectraKeys.lambda_array:
            return self.lambda_bin_margin
        else:
            raise NotImplementedError(f"axis type {axis_type} not implemented!")

    def checked_range(self, left=0, right=0):
        """this method makes sure that the left and right values stay within the maximum range of the data
        for the current axis selected"""
        o_get = Get(parent=self.parent)
        x_axis_type_selected = o_get.bin_x_axis_selected()
        x_axis = self.parent.time_spectra[x_axis_type_selected]

        if left < x_axis[0]:
            left = x_axis[0]

        if right >= x_axis[-1]:
            right = x_axis[-1]

        index_clean_left_value = get_value_of_closest_match(array_to_look_for=x_axis,
                                                            value=left,
                                                            left_margin=True)
        index_clean_right_value = get_value_of_closest_match(array_to_look_for=x_axis,
                                                             value=right,
                                                             left_margin=False)

        clean_left_value = x_axis[index_clean_left_value]
        if clean_left_value < left:
            clean_left_value = x_axis[index_clean_left_value + 1]

        clean_right_value = x_axis[index_clean_right_value]
        if clean_right_value > right:
            clean_right_value = x_axis[index_clean_right_value - 1]

        return clean_left_value, clean_right_value



import pytest

from maverick.combine_bin_cli import CombineCLI


class TestMaverickCombineCLI:

    def test_wrong_number_of_files(self):
        algorithm = 'median'
        output_folder = '/Users/j35/Desktop/test_maverick_cli'
        input_folders = ['/Volumes/Buffalo/IPTS/IPTS-30023-Matteo-Simon/scan4/Run_56305/',
                         '/Volumes/Buffalo/IPTS/IPTS-30023-Matteo-Simon/scan4/Run_56307/']
        str_input_folders = " ".join(input_folders)

        with pytest.raises(ValueError) as exc:
            CombineCLI(list_of_folders=input_folders)

    def test_not_enough_folders(self):
        algorithm = 'median'
        output_folder = '/Users/j35/Desktop/test_maverick_cli'

        # only 1 folder
        input_folders = ['/Volumes/Buffalo/IPTS/IPTS-30023-Matteo-Simon/scan4/Run_56305/']
        str_input_folders = " ".join(input_folders)

        with pytest.raises(ValueError) as exc:
            CombineCLI(list_of_folders=input_folders)

        # 0 folder given
        with pytest.raises(ValueError) as exc:
            CombineCLI(list_of_folders=[])



from qtpy.QtWidgets import QDialog
import os

from .. import load_ui
from .session_handler import SessionHandler
from ..utilities.get import Get
from ..utilities.check import Check
# from .load_previous_session_launcher_multiple_choice import LoadPreviousSessionLauncherMultipleChoice


class LoadPreviousSessionLauncher(QDialog):

    def __init__(self, parent=None, config=None):
        self.parent = parent
        QDialog.__init__(self, parent=parent)
        ui_full_path = os.path.join(os.path.dirname(__file__),
                                    os.path.join('ui',
                                                 'ui_load_previous_session.ui'))
        self.ui = load_ui(ui_full_path, baseinstance=self)
        self.setWindowTitle("Load previous session?")
        self.ui.pushButton.setFocus(True)

    def yes_clicked(self):
        self.close()
        o_session = SessionHandler(parent=self.parent)
        o_get = Get(parent=self.parent)
        full_config_file_name = o_get.automatic_config_file_name()
        o_session.load_from_file(config_file_name=full_config_file_name)

        # list_tabs_to_load = o_session.get_tabs_to_load()
        # if len(list_tabs_to_load) < 2:
        #     o_session.load_to_ui(tabs_to_load=list_tabs_to_load)
        #     self.parent.loading_from_config = False
        # else:
        #     load_session_ui = LoadPreviousSessionLauncherMultipleChoice(parent=self.parent,
        #                                                                 list_tabs_to_load=list_tabs_to_load)
        #     load_session_ui.show()
        # self.parent.check_log_file_size()
        o_check = Check(parent=self.parent)
        o_check.log_file_size()

        self.parent.check_combine_widgets()

    def no_clicked(self):
        self.close()
        o_check = Check(parent=self.parent)
        o_check.log_file_size()

    def reject(self):
        self.no_clicked()

    def closeEvent(self, event):
        self.close()



import os
import logging

from .utilities.get import Get
from .session.load_previous_session_launcher import LoadPreviousSessionLauncher
from .session.session_handler import SessionHandler


class EventHandler:

    def __init__(self, parent=None):
        self.parent = parent
        self.logger = logging.getLogger("maverick")

    def automatically_load_previous_session(self):
        o_get = Get(parent=self.parent)
        full_config_file_name = o_get.automatic_config_file_name()
        if os.path.exists(full_config_file_name):
            load_session_ui = LoadPreviousSessionLauncher(parent=self.parent)
            load_session_ui.show()
        else:
            o_session = SessionHandler(parent=self.parent)
            self.session = o_session.session



from pathlib import Path
from astropy.io import fits
import numpy as np
import os
import shutil
import glob


class FileHandler(object):

    @classmethod
    def get_list_of_folders(cls, top_folder):
        full_list = glob.glob(top_folder + "/*")
        full_list.sort()
        list_folders = []
        for _entry in full_list:
            if os.path.isdir(_entry):
                list_folders.append(_entry)
        return list_folders

    @classmethod
    def get_list_of_files(cls, folder):
        full_list = glob.glob(folder + "/*")
        full_list.sort()
        return full_list

    @classmethod
    def get_list_of_tif(cls, folder):
        full_list = glob.glob(folder + '/*.tif*')
        full_list.sort()
        return full_list

    @classmethod
    def get_parent_folder(cls, full_folder):
        folder_parts = Path(full_folder).parts
        return folder_parts[-2]

    @classmethod
    def get_parent_path(cls, full_path):
        full_path = Path(full_path)
        return full_path.parent

    @classmethod
    def get_base_filename(cls, full_file_name):
        return Path(full_file_name).name

    @staticmethod
    def get_file_extension(file_name):
        ext = Path(file_name).suffix
        return ext

    @classmethod
    def retrieve_ascii_contain(cls, full_file_name):
        file_contain = []
        with open(full_file_name) as f:
            file_contain = f.read()
        return file_contain

    @classmethod
    def cleanup_list_of_files(cls, list_of_files=[], base_number=5):
        '''Will only keep the files that have the same number of character as
        the first n files'''
        if list_of_files == []:
            return []

        len_base_files = []
        for _file in list_of_files[0: base_number]:
            len_base_files.append(len(_file))

        # make sure all the length of the base number files match
        set_len_base_files = set(len_base_files)
        if len(set_len_base_files) > 1:
            raise ValueError("Format Input File Do Not Match!")

        len_file = len_base_files[0]
        final_list = []
        for _file in list_of_files:
            if len(_file) == len_file:
                final_list.append(_file)

        return final_list

    @classmethod
    def make_fits(cls, data=[], filename=''):
        fits.writeto(filename, data, overwrite=True)

    @classmethod
    def make_ascii_file(cls, metadata=[], data=[], output_file_name='', sep=','):
        f = open(output_file_name, 'w')
        for _meta in metadata:
            _line = _meta + "\n"
            f.write(_line)

        if len(np.shape(data)) > 1:
            for _data in data:
                _str_data = [str(_value) for _value in _data]
                _line = sep.join(_str_data) + "\n"
                f.write(_line)
        else:
            _str_data = [str(_value) + "\n" for _value in data]
            for _data in _str_data:
                f.write(_data)

        f.close()

    @staticmethod
    def make_or_reset_folder(folder_name):
        if os.path.exists(folder_name):
            shutil.rmtree(folder_name)
        os.makedirs(folder_name)


def read_ascii(filename=''):
    '''return contain of an ascii file'''
    with open(filename, 'r') as f:
        text = f.read()
    return text


def write_ascii(text="", filename=''):
    with open(filename, 'w') as f:
        f.write(text)



from qtpy.QtWidgets import QApplication, QMainWindow
import sys
import os
import logging
import warnings

warnings.filterwarnings("ignore")

from .utilities.get import Get
from .utilities.config_handler import ConfigHandler
from .utilities import TimeSpectraKeys, BinAutoMode
from .utilities.time_spectra import TimeSpectraLauncher
from .event_hander import EventHandler
from .session import session
from .session.session_handler import SessionHandler
from .session import SessionKeys
from .initialization import Initialization
from .utilities.check import Check
from .combine.event_handler import EventHandler as CombineEventHandler
from .bin.event_hander import EventHandler as BinEventHandler
from .bin.manual_event_handler import ManualEventHandler as BinManualEventHandler
from .bin.auto_event_handler import AutoEventHandler as BinAutoEventHandler
from .bin.preview_full_bin_axis import PreviewFullBinAxis
from .bin.statistics import Statistics
from .bin.settings import Settings as BinSettings
from .bin.manual_right_click import ManualRightClick
from maverick.export.export_images import ExportImages
from maverick.export.export_bin_table import ExportBinTable
from .log.log_launcher import LogLauncher

from . import load_ui


class MainWindow(QMainWindow):
    session = session  # dictionary that will keep record of the entire UI and used to load and save the session
    log_id = None  # ui id of the log QDialog
    version = None   # current version of application

    # raw_data_folders = {'full_path_to_folder1': {'data': [image1, image2, image3...],
    #                                              'list_files': [file1, file2, file3,...],
    #                                              'nbr_files': 0,
    #                                              },
    #                     'full_path_to_folder2': {'data': [image1, image2, image3...],
    #                                              'list_files': [file1, file2, file3,...],
    #                                              'nbr_files': 0,
    #                                              },
    #                     ....
    #                    }
    raw_data_folders = None  # dictionary of data for each of the folders

    # combine_data = [image1, image2, image3...]
    combine_data = None

    # time spectra file and arrays
    time_spectra = {TimeSpectraKeys.file_name: None,
                    TimeSpectraKeys.tof_array: None,
                    TimeSpectraKeys.lambda_array: None,
                    TimeSpectraKeys.file_index_array: None}

    linear_bins = {TimeSpectraKeys.tof_array: None,
                   TimeSpectraKeys.file_index_array: None,
                   TimeSpectraKeys.lambda_array: None}

    log_bins = {TimeSpectraKeys.tof_array: None,
                TimeSpectraKeys.file_index_array: None,
                TimeSpectraKeys.lambda_array: None}

    # each will be a dictionaries of ranges
    # ex: TimeSpectraKeys.tof_array = {0: [1],
    #                                  1: [2,6],
    #                                  3: [7,8,9,10], ...}
    manual_bins = {TimeSpectraKeys.tof_array: None,
                   TimeSpectraKeys.file_index_array: None,
                   TimeSpectraKeys.lambda_array: None}

    # dictionary that will record the range for each bin
    # {0: [0, 3], 1: [1, 10], ...}
    manual_snapping_indexes_bins = None

    # list of rows selected by each of the linear and log bins
    linear_bins_selected = None
    log_bins_selected = None

    # use to preview the full axis
    # ex: [1,2,3,4,5,6,7] or [0.1, 0.2, 0.4, 0.8, 1.6....]
    full_bin_axis_requested = None

    # profile signal (displayed on the top right of combine and bin tab)
    # 1D array
    profile_signal = None

    # pyqtgraph view
    combine_image_view = None  # combine image view id - top right plot
    combine_profile_view = None  # combine profile plot view id - bottom right plot
    bin_profile_view = None  # bin profile
    combine_roi_item_id = None  # pyqtgraph item id of the roi (combine tab)
    combine_file_index_radio_button = None  # in combine view
    tof_radio_button = None  # in combine view
    lambda_radio_button = None  # in combine view
    live_combine_image = None  # live combine image used by ROI

    # matplotlib plot
    statistics_plot = None  # matplotlib plot

    # dictionary of all the bins pg item
    # {0: pg.regionitem1,
    #  2: pg.regionitem2,
    #  ...
    # }
    dict_of_bins_item = None

    # list of manual bins.
    # using a list because any of the bin can be removed by the user
    list_of_manual_bins_item = []

    current_auto_bin_rows_highlighted = []

    # stats currently displayed in the bin stats table
    # {StatisticsName.mean: {Statistics.full: [],
    #                        Statistics.roi: [],
    #                        },
    # StatisticsName.median: ....
    #  }
    current_stats = None

    def __init__(self, parent=None):
        """
        Initialization
        Parameters
        ----------
        """
        super(MainWindow, self).__init__(parent)
        self.ui = load_ui('mainWindow.ui', baseinstance=self)
        self.initialization()
        self.setup()
        self.setWindowTitle(f"maverick - v{self.version}")

    def initialization(self):
        o_init = Initialization(parent=self)
        o_init.all()

    def setup(self):
        """
        This is taking care of
            - initializing the session dict
            - setting up the logging
            - retrieving the config file
            - loading or not the previous session
        """
        o_config = ConfigHandler(parent=self)
        o_config.load()

        current_folder = None
        if self.config['debugging']:
            list_homepath = self.config['homepath']
            for _path in list_homepath:
                if os.path.exists(_path):
                    current_folder = _path
            if current_folder is None:
                current_folder = os.path.expanduser('~')
        else:
            current_folder = os.path.expanduser('~')
        self.session[SessionKeys.top_folder] = current_folder

        o_get = Get(parent=self)
        log_file_name = o_get.log_file_name()
        version = Get.version()
        self.version = version
        self.log_file_name = log_file_name
        logging.basicConfig(filename=log_file_name,
                            filemode='a',
                            format='[%(levelname)s] - %(asctime)s - %(message)s',
                            level=logging.INFO)
        logger = logging.getLogger("maverick")
        logger.info("*** Starting a new session ***")
        logger.info(f" Version: {version}")

        o_event = EventHandler(parent=self)
        o_event.automatically_load_previous_session()

    # Menu
    def session_load_clicked(self):
        o_session = SessionHandler(parent=self)
        o_session.load_from_file()
        o_session.load_to_ui()

    def session_save_clicked(self):
        o_session = SessionHandler(parent=self)
        o_session.save_from_ui()
        o_session.save_to_file()

    def help_log_clicked(self):
        LogLauncher(parent=self)

    # combine events
    def check_combine_widgets(self):
        o_event = CombineEventHandler(parent=self)
        o_event.check_widgets()

    def combine_bin_tab_changed(self, new_tab_index):
        if new_tab_index == 1:  # bin
            o_event = BinEventHandler(parent=self)
            o_event.entering_tab()
            self.update_statistics()

    def select_top_folder_button_clicked(self):
        o_event = CombineEventHandler(parent=self)
        o_event.select_top_folder()

    def refresh_table_clicked(self):
        o_event = CombineEventHandler(parent=self)
        o_event.refresh_table_clicked()

    def radio_buttons_of_folder_changed(self):
        self.ui.setEnabled(False)
        o_event = CombineEventHandler(parent=self)
        o_event.update_list_of_folders_to_use()
        o_event.combine_folders()
        o_event.display_profile()
        o_event.check_widgets()
        self.ui.setEnabled(True)

    def time_spectra_preview_clicked(self):
        TimeSpectraLauncher(parent=self)

    def combine_algorithm_changed(self):
        o_get = Get(parent=self)
        list_working_folders = o_get.list_of_folders_to_use()
        if list_working_folders == []:
            return

        o_event = CombineEventHandler(parent=self)
        o_event.combine_algorithm_changed()
        o_event.display_profile()

    def combine_instrument_settings_changed(self):
        if self.combine_data is None:
            return
        o_event = CombineEventHandler(parent=self)
        o_event.update_list_of_folders_to_use(force_recalculation_of_time_spectra=True)
        o_event.combine_folders()
        o_event.display_profile()

    def combine_xaxis_changed(self):
        o_event = CombineEventHandler(parent=self)
        o_event.display_profile()

    def combine_roi_changed(self):
        o_event = CombineEventHandler(parent=self)
        o_event.combine_roi_changed()
        o_event.display_profile()

    # bin events
    def bin_xaxis_changed(self):
        o_event = BinEventHandler(parent=self)
        o_event.bin_axis_changed()

    def bin_auto_manual_tab_changed(self, new_tab_index):
        o_event = BinEventHandler(parent=self)
        o_event.bin_auto_manual_tab_changed(new_tab_index)
        self.update_statistics()

    # - auto mode
    def bin_auto_log_linear_radioButton_changed(self):
        o_event = BinAutoEventHandler(parent=self)
        o_event.bin_auto_radioButton_clicked()
        self.update_statistics()

    def bin_auto_log_file_index_changed(self):
        o_event = BinAutoEventHandler(parent=self)
        o_event.bin_auto_log_changed(source_radio_button=TimeSpectraKeys.file_index_array)
        self.update_statistics()

    def bin_auto_log_tof_changed(self):
        o_event = BinAutoEventHandler(parent=self)
        o_event.bin_auto_log_changed(source_radio_button=TimeSpectraKeys.tof_array)
        self.update_statistics()

    def bin_auto_log_lambda_changed(self):
        o_event = BinAutoEventHandler(parent=self)
        o_event.bin_auto_log_changed(source_radio_button=TimeSpectraKeys.lambda_array)
        self.update_statistics()

    def bin_auto_linear_file_index_changed(self):
        o_event = BinAutoEventHandler(parent=self)
        o_event.clear_selection(auto_mode=BinAutoMode.linear)
        o_event.bin_auto_linear_changed(source_radio_button=TimeSpectraKeys.file_index_array)
        self.update_statistics()

    def bin_auto_linear_tof_changed(self):
        o_event = BinAutoEventHandler(parent=self)
        o_event.clear_selection(auto_mode=BinAutoMode.linear)
        o_event.bin_auto_linear_changed(source_radio_button=TimeSpectraKeys.tof_array)
        self.update_statistics()

    def bin_auto_linear_lambda_changed(self):
        o_event = BinAutoEventHandler(parent=self)
        o_event.clear_selection(auto_mode=BinAutoMode.linear)
        o_event.bin_auto_linear_changed(source_radio_button=TimeSpectraKeys.lambda_array)
        self.update_statistics()

    def auto_log_radioButton_changed(self):
        o_event = BinAutoEventHandler(parent=self)
        o_event.clear_selection(auto_mode=BinAutoMode.log)
        o_event.auto_log_radioButton_changed()
        self.update_statistics()

    def auto_linear_radioButton_changed(self):
        o_event = BinAutoEventHandler(parent=self)
        o_event.clear_selection(auto_mode=BinAutoMode.linear)
        o_event.auto_linear_radioButton_changed()
        self.update_statistics()

    def auto_table_use_checkbox_changed(self, state, row):
        o_event = BinAutoEventHandler(parent=self)
        state = True if state == 2 else False
        o_event.use_auto_bin_state_changed(row=row, state=state)
        self.bin_auto_table_selection_changed()
        self.update_statistics()

    def bin_auto_hide_empty_bins(self):
        o_event = BinAutoEventHandler(parent=self)
        o_event.update_auto_table()

    def bin_auto_visualize_axis_generated_button_clicked(self):
        o_preview = PreviewFullBinAxis(parent=self)
        o_preview.show()

    def bin_auto_table_right_clicked(self, position):
        o_event = BinAutoEventHandler(parent=self)
        o_event.auto_table_right_click(position=position)

    def bin_auto_table_selection_changed(self):
        o_event = BinAutoEventHandler(parent=self)
        o_event.auto_table_selection_changed()

    def mouse_moved_in_combine_image_preview(self):
        """Mouse moved in the combine pyqtgraph image preview (top right)"""
        pass

    # - manual mode
    def bin_manual_add_bin_clicked(self):
        o_event = BinManualEventHandler(parent=self)
        o_event.add_bin()
        self.update_statistics()

    def bin_manual_populate_table_with_auto_mode_bins_clicked(self):
        o_event = BinManualEventHandler(parent=self)
        o_event.clear_all_items()
        o_event.populate_table_with_auto_mode()
        self.update_statistics()

    def bin_manual_region_changed(self, item_id):
        o_event = BinManualEventHandler(parent=self)
        o_event.bin_manually_moved(item_id=item_id)
        self.update_statistics()

    def bin_manual_region_changing(self, item_id):
        o_event = BinManualEventHandler(parent=self)
        o_event.bin_manually_moving(item_id=item_id)

    def bin_manual_table_right_click(self, position):
        o_event = ManualRightClick(parent=self)
        o_event.manual_table_right_click()

    # - statistics
    def update_statistics(self):
        o_stat = Statistics(parent=self)
        o_stat.update()
        o_stat.plot_statistics()

    def bin_statistics_comboBox_changed(self):
        o_stat = Statistics(parent=self)
        o_stat.plot_statistics()

    def bin_settings_clicked(self):
        o_bin = BinSettings(parent=self)
        o_bin.show()

    # export images
    def export_combined_and_binned_images_clicked(self):
        o_export = ExportImages(parent=self)
        o_export.run()

    def bin_export_table_pushButton_clicked(self):
        o_export = ExportBinTable(parent=self)
        o_export.run()

    def closeEvent(self, event):
        o_session = SessionHandler(parent=self)
        o_session.save_from_ui()
        o_session.automatic_save()

        o_event = Check(parent=self)
        o_event.log_file_size()

        logging.info(" #### Leaving maverick ####")
        self.close()


def main(args):
    app = QApplication(args)
    app.setStyle("Fusion")
    app.aboutToQuit.connect(clean_up)
    app.setApplicationDisplayName("maverick")
    # app.setWindowIcon(PyQt4.QtGui.QIcon(":/icon.png"))
    application = MainWindow()
    application.show()
    sys.exit(app.exec_())


def clean_up():
    app = QApplication.instance()
    app.closeAllWindows()



import os
from setuptools import setup, find_packages
import versioneer  # https://github.com/warner/python-versioneer

THIS_DIR = os.path.dirname(__file__)


def read_requirements_from_file(filepath):
    '''Read a list of requirements from the given file and split into a
    list of strings. It is assumed that the file is a flat
    list with one requirement per line.
    :param filepath: Path to the file to read
    :return: A list of strings containing the requirements
    '''
    with open(filepath, 'r') as req_file:
        return req_file.readlines()


setup_args = dict(
        install_requires=read_requirements_from_file(
                os.path.join(
                        THIS_DIR,
                        'requirements.txt')),
        # tests_require=read_requirements_from_file(
        #     os.path.join(
        #         THIS_DIR,
        #         'requirements-dev.txt'))
)

setup(name="maverick",
      #      version=versioneer.get_version(),
      version="not_used",
      description="TOF binning and combining utility",
      author="Jean Bilheux",
      author_email="bilheuxjm@ornl.gov",
      url="http://github.com/ornlneutronimaging/maverick",
      long_description="""Should have a longer description""",
      license="The MIT License (MIT)",
      packages=find_packages(),
      package_data={'': ['*.ui', '*.png', '*.qrc', '*.json']},
      install_requires=setup_args["install_requires"],
      setup_requires=[],
      )



from qtpy.QtWidgets import QFileDialog, QApplication
import json
import logging
import copy

from ..utilities.status_message_config import StatusMessageStatus, show_status_message
from ..combine.event_handler import EventHandler as CombineEventHandler

from . import SessionKeys
from . import session as default_session
from ..utilities.get import Get
from ..utilities import CombineAlgorithm, BinMode


class SessionHandler:

    logger = None   # customized logging for maverick

    config_file_name = ""
    load_successful = True

    session = None

    def __init__(self, parent=None):
        self.logger = logging.getLogger("maverick")
        self.logger.info("-> Saving current session before leaving the application")
        self.parent = parent

    def save_from_ui(self):
        pass

    def load_to_ui(self):
        if not self.load_successful:
            return

        self.logger.info(f"Automatic loading of session")
        session = copy.deepcopy(self.parent.session)
        self.logger.info(f"session -> {session}")

        # combine
        self.parent.ui.top_folder_label.setText(session[SessionKeys.top_folder])
        self.parent.ui.distance_source_detector_doubleSpinBox.setValue(session[SessionKeys.distance_source_detector])
        self.parent.ui.detector_offset_spinBox.setValue(session[SessionKeys.detector_offset])
        self.parent.ui.combine_sample_position_doubleSpinBox.setValue(session[SessionKeys.sample_position])
        o_combine_event = CombineEventHandler(parent=self.parent)
        o_combine_event.reset_data()
        self.parent.session = session
        o_combine_event.populate_list_of_folders_to_combine()
        o_combine_event.update_list_of_folders_to_use()

        combine_algorithm = session.get(SessionKeys.combine_algorithm, CombineAlgorithm.mean)
        if combine_algorithm == CombineAlgorithm.mean:
            self.parent.ui.combine_mean_radioButton.setChecked(True)
        elif combine_algorithm == CombineAlgorithm.median:
            self.parent.ui.combine_median_radioButton.setChecked(True)
        else:
            raise NotImplementedError("Combine method not implemented!")

        if not (SessionKeys.combine_roi in session.keys()):
            self.parent.session[SessionKeys.combine_roi] = default_session[SessionKeys.combine_roi]

        o_combine_event.combine_folders()
        o_combine_event.display_profile()

        bin_mode = session.get(SessionKeys.bin_mode, BinMode.auto)
        if bin_mode == BinMode.auto:
            self.parent.bin_tabWidget.setCurrentIndex(0)
        elif bin_mode == BinMode.manual:
            self.parent.bin_tabWidget.setCurrentIndex(1)
        else:
            raise NotImplementedError("Auto bin mode not implemented!")

        bin_algorithm = session.get(SessionKeys.bin_algorithm, CombineAlgorithm.mean)
        self.parent.session[SessionKeys.bin_algorithm] = bin_algorithm

    def automatic_save(self):
        self.logger.info(self.parent.session)
        o_get = Get(parent=self.parent)
        full_config_file_name = o_get.automatic_config_file_name()
        self.save_to_file(config_file_name=full_config_file_name)

    def save_to_file(self, config_file_name=None):
        if config_file_name is None:
            config_file_name = QFileDialog.getSaveFileName(self.parent,
                                                           caption="Select session file name ...",
                                                           directory=self.parent.session[SessionKeys.top_folder],
                                                           filter="session (*.json)",
                                                           initialFilter="session")

            QApplication.processEvents()
            config_file_name = config_file_name[0]

        if config_file_name:
            output_file_name = config_file_name
            session = self.parent.session

            with open(output_file_name, 'w') as json_file:
                json.dump(session, json_file)

            show_status_message(parent=self.parent,
                                message=f"Session saved in {config_file_name}",
                                status=StatusMessageStatus.ready,
                                duration_s=10)
            self.logger.info(f"Saving configuration into {config_file_name}")

    def load_from_file(self, config_file_name=None):
        if config_file_name is None:
            config_file_name = QFileDialog.getOpenFileName(self.parent,
                                                           directory=self.parent.session[SessionKeys.top_folder],
                                                           caption="Select session file ...",
                                                           filter="session (*.json)",
                                                           initialFilter="session")
            QApplication.processEvents()
            config_file_name = config_file_name[0]

        if config_file_name:
            config_file_name = config_file_name
            self.config_file_name = config_file_name
            show_status_message(parent=self.parent,
                                message=f"Loading {config_file_name} ...",
                                status=StatusMessageStatus.ready)

            with open(config_file_name, "r") as read_file:
                session = json.load(read_file)
                o_get = Get(parent=self.parent)
                maverick_current_version = o_get.version()
                if session[SessionKeys.version] == maverick_current_version:
                    self.parent.session = session
                    self.load_to_ui()
                    self.logger.info(f"Loaded from {config_file_name}")
                    self.load_successful = True
                else:
                    self.logger.info(f"Session file is out of date!")
                    self.logger.info(f"-> expected version: {maverick_current_version}")
                    self.logger.info(f"-> session version: {session[SessionKeys.version]}")
                    self.load_successful = False

                if not self.load_successful:
                    show_status_message(parent=self.parent,
                                        message=f"{config_file_name} not loaded! (check log for more information)",
                                        status=StatusMessageStatus.ready,
                                        duration_s=10)

                else:
                    self.logger.info(f"Loading {config_file_name} ... Done!")
                    show_status_message(parent=self.parent,
                                        message=f"Loaded {config_file_name} ... Done!",
                                        status=StatusMessageStatus.ready,
                                        duration_s=10)

        else:
            self.load_successful = False
            show_status_message(parent=self.parent,
                                message=f"{config_file_name} not loaded! (check log for more information)",
                                status=StatusMessageStatus.ready,
                                duration_s=10)



import numpy as np

from ..utilities.get import Get
from ..utilities import BinMode, BinAutoMode, TimeSpectraKeys, BinAlgorithm
from ..utilities.string import format_str
from ..utilities.table_handler import TableHandler
from ..session import SessionKeys

from . import TO_MICROS_UNITS, TO_ANGSTROMS_UNITS
from . import StatisticsName, StatisticsRegion


class Statistics:

    def __init__(self, parent=None):
        self.parent = parent

    def update(self):

        o_get = Get(parent=self.parent)

        # check if we are looking for the auto or the manual bins
        bin_mode = o_get.bin_mode()

        if bin_mode == BinMode.auto:

            bin_auto_mode = o_get.bin_auto_mode()
            if bin_auto_mode == BinAutoMode.linear:
                list_bins = self.parent.linear_bins
            else:
                list_bins = self.parent.log_bins

        else:

            list_bins = self.parent.manual_bins

        # if no bins to display, stop here
        if list_bins[TimeSpectraKeys.file_index_array] is None:
            return

        o_table = TableHandler(table_ui=self.parent.ui.statistics_tableWidget)
        o_table.remove_all_rows()

        file_index_bins = list_bins[TimeSpectraKeys.file_index_array]
        tof_bins = list_bins[TimeSpectraKeys.tof_array]
        lambda_bins = list_bins[TimeSpectraKeys.lambda_array]

        mean_array_full = []
        median_array_full = []
        std_array_full = []
        min_array_full = []
        max_array_full = []

        mean_array_roi = []
        median_array_roi = []
        std_array_roi = []
        min_array_roi = []
        max_array_roi = []

        _row = 0
        for _bin_index, _bin in enumerate(file_index_bins):

            o_table.insert_empty_row(row=_row)

            o_table.insert_item(row=_row,
                                column=0,
                                value=str(_row),
                                editable=False)

            if _bin == []:

                mean_array_full.append(np.nan)
                median_array_full.append(np.nan)
                std_array_full.append(np.nan)
                min_array_full.append(np.nan)
                max_array_full.append(np.nan)

                mean_array_roi.append(np.nan)
                median_array_roi.append(np.nan)
                std_array_roi.append(np.nan)
                min_array_roi.append(np.nan)
                max_array_roi.append(np.nan)

                list_runs_formatted = "N/A"
                list_tof_formatted = "N/A"
                list_lambda_formatted = "N/A"
                str_mean = "N/A"
                str_median = "N/A"
                str_std = "N/A"
                str_min = "N/A"
                str_max = "N/A"

            else:

                list_runs = file_index_bins[_bin_index]
                list_runs_formatted = format_str(list_runs,
                                                 format_str="{:d}",
                                                 factor=1,
                                                 data_type=TimeSpectraKeys.file_index_array)

                list_tof = tof_bins[_bin_index]
                list_tof_formatted = format_str(list_tof,
                                                format_str="{:.2f}",
                                                factor=TO_MICROS_UNITS,
                                                data_type=TimeSpectraKeys.tof_array)

                list_lambda = lambda_bins[_bin_index]
                list_lambda_formatted = format_str(list_lambda,
                                                   format_str="{:.3f}",
                                                   factor=TO_ANGSTROMS_UNITS,
                                                   data_type=TimeSpectraKeys.lambda_array)

                # calculate statistics
                _data_dict = self.extract_data_for_this_bin(list_runs=list_runs)

                full_image = _data_dict['full_image']
                roi_of_image = _data_dict['roi_of_image']

                # mean
                full_mean = np.mean(full_image)
                roi_mean = np.mean(roi_of_image)
                mean_array_full.append(full_mean)
                mean_array_roi.append(roi_mean)
                str_mean = f"{full_mean:.3f} ({roi_mean:.3f})"

                # median
                full_median = np.median(full_image)
                roi_median = np.median(roi_of_image)
                median_array_full.append(full_median)
                median_array_roi.append(roi_median)
                str_median = f"{full_median:.3f} ({roi_median:.3f})"

                # std
                full_std = np.std(full_image)
                roi_std = np.std(roi_of_image)
                std_array_full.append(full_std)
                std_array_roi.append(roi_std)
                str_std = f"{full_std:.3f} ({roi_std:.3f})"

                # min
                full_min = np.min(full_image)
                roi_min = np.min(roi_of_image)
                min_array_full.append(full_min)
                min_array_roi.append(roi_min)
                str_min = f"{full_min:.3f} ({roi_min:.3f})"

                # max
                full_max = np.max(full_image)
                roi_max = np.max(roi_of_image)
                max_array_full.append(full_max)
                max_array_roi.append(roi_max)
                str_max = f"{full_max:.3f} ({roi_max:.3f})"

            o_table.insert_item(row=_row,
                                column=1,
                                value=list_runs_formatted,
                                editable=False)

            o_table.insert_item(row=_row,
                                column=2,
                                value=list_tof_formatted,
                                editable=False)

            o_table.insert_item(row=_row,
                                column=3,
                                value=list_lambda_formatted,
                                editable=False)

            o_table.insert_item(row=_row,
                                column=4,
                                value=str_mean,
                                editable=False)

            o_table.insert_item(row=_row,
                                column=5,
                                value=str_median,
                                editable=False)

            o_table.insert_item(row=_row,
                                column=6,
                                value=str_std,
                                editable=False)

            o_table.insert_item(row=_row,
                                column=7,
                                value=str_min,
                                editable=False)

            o_table.insert_item(row=_row,
                                column=8,
                                value=str_max,
                                editable=False)

            _row += 1

        current_stats = {StatisticsName.mean: {StatisticsRegion.full: mean_array_full,
                                               StatisticsRegion.roi: mean_array_roi},
                         StatisticsName.median: {StatisticsRegion.full: median_array_full,
                                                 StatisticsRegion.roi: median_array_roi},
                         StatisticsName.std: {StatisticsRegion.full: std_array_full,
                                              StatisticsRegion.roi: std_array_roi},
                         StatisticsName.min: {StatisticsRegion.full: min_array_full,
                                              StatisticsRegion.roi: min_array_roi},
                         StatisticsName.max: {StatisticsRegion.full: max_array_full,
                                              StatisticsRegion.roi: max_array_roi}}

        self.parent.current_stats = current_stats

    def extract_data_for_this_bin(self, list_runs=None):
        """
        this method isolate the data of only the runs of the corresponding runs, and only for the ROI selected

        :param list_runs:
        :return:
        """
        # retrieve statistics
        [x0, y0, width, height] = self.parent.session[SessionKeys.combine_roi]

        data_to_work_with = []
        for _run_index in list_runs:
            data_to_work_with.append(self.parent.combine_data[_run_index])

        region_to_work_with = [_data[y0: y0+height, x0: x0+width] for _data in data_to_work_with]

        # how to add images
        o_get = Get(parent=self.parent)
        bin_method = o_get.bin_add_method()
        if bin_method == BinAlgorithm.mean:
            full_image_to_work_with = np.mean(data_to_work_with, axis=0)
            roi_image_to_work_with = np.mean(region_to_work_with, axis=0)
        elif bin_method == BinAlgorithm.median:
            full_image_to_work_with = np.median(data_to_work_with, axis=0)
            roi_image_to_work_with = np.median(region_to_work_with, axis=0)
        else:
            raise NotImplementedError("this method of adding the binned images is not supported!")

        return {'full_image': full_image_to_work_with,
                'roi_of_image': roi_image_to_work_with}

    def plot_statistics(self):
        if self.parent.current_stats is None:
            self.parent.statistics_plot.ax1.clear()
            return

        o_get = Get(parent=self.parent)
        stats_requested = o_get.bin_statistics_plot_requested()

        stat_data_dict = self.parent.current_stats[stats_requested]
        full_array = stat_data_dict[StatisticsRegion.full]
        roi_array = stat_data_dict[StatisticsRegion.roi]

        self.parent.statistics_plot.ax1.clear()
        self.parent.statistics_plot.ax1.plot(full_array, '.', label='full image')
        self.parent.statistics_plot.ax1.plot(roi_array, '+', label='roi only')

        self.parent.statistics_plot.ax1.set_xlabel(f"Bin #")
        self.parent.statistics_plot.ax1.set_ylabel(stats_requested)
        self.parent.statistics_plot.ax1.legend(loc='upper right')

        self.parent.statistics_plot.draw()



from pathlib import Path
from ..bin import TO_ANGSTROMS_UNITS, TO_MICROS_UNITS


def create_bin_tab_output_file_name(folder=None,
                                    bin_name=None):
    bin_name += "_bin_table.json"
    full_file_name = Path(folder) / bin_name
    return full_file_name


def create_output_file_name(folder=None,
                            bin_index=0,
                            sample_position=0,
                            list_file_index=None,
                            list_tof=None,
                            list_lambda=None):

    sample_position = str(sample_position).replace(".", "_")
    bin_index = f"{bin_index:02d}"

    if len(list_file_index) == 1:
        str_file = f"file_#{list_file_index[0]}"

        str_tof_value = convert_to_user_friendly_micros_units(list_tof[0])
        str_tof = f"tof_{str_tof_value}_micros"

        str_lambda_value = convert_to_user_friendly_angstroms_units(list_lambda[0])
        str_lambda = f"lambda_{str_lambda_value}_angstroms"

    elif len(list_file_index) == 2:
        str_file = f"file_{list_file_index[0]}_#{list_file_index[1]}"

        from_str_tof_value = convert_to_user_friendly_micros_units(list_tof[0])
        to_str_tof_value = convert_to_user_friendly_micros_units(list_tof[1])
        str_tof = f"from_tof_{from_str_tof_value}_to_{to_str_tof_value}_micros"

        from_str_lambda_value = convert_to_user_friendly_angstroms_units(list_lambda[0])
        to_str_lambda_value = convert_to_user_friendly_angstroms_units(list_lambda[1])
        str_lambda = f"from_lambda_{from_str_lambda_value}_to_{to_str_lambda_value}_angstroms"

    else:
        str_file = f"from_file_#{list_file_index[0]}_to_file_#{list_file_index[-1]}"

        from_str_tof_value = convert_to_user_friendly_micros_units(list_tof[0])
        to_str_tof_value = convert_to_user_friendly_micros_units(list_tof[-1])
        str_tof = f"from_tof_{from_str_tof_value}_to_{to_str_tof_value}_micros"

        from_str_lambda_value = convert_to_user_friendly_angstroms_units(list_lambda[0])
        to_str_lambda_value = convert_to_user_friendly_angstroms_units(list_lambda[-1])
        str_lambda = f"from_lambda_{from_str_lambda_value}_to_{to_str_lambda_value}_angstroms"

    str_bin = f"bin#_{bin_index}"
    str_sample_position = f"sample_position_{sample_position}_mm"

    file_name = f"image_{str_bin}_{str_sample_position}_{str_file}_{str_tof}_{str_lambda}.tiff"

    full_file_name = Path(folder) / file_name

    return full_file_name


def convert_to_user_friendly_angstroms_units(value):
    angstroms_str = f"{value * TO_ANGSTROMS_UNITS:.2f}"
    new_angstroms_str = angstroms_str.replace(".", "_")
    return new_angstroms_str


def convert_to_user_friendly_micros_units(value):
    tof_str = f"{value * TO_MICROS_UNITS:.2f}"
    new_tof_str = tof_str.replace(".", "_")
    return new_tof_str



from qtpy.QtWidgets import QProgressBar, QVBoxLayout, QHBoxLayout, QRadioButton
from qtpy.QtWidgets import QSpacerItem, QSizePolicy, QWidget
from qtpy.QtGui import QIcon
from pyqtgraph.dockarea import DockArea, Dock
import pyqtgraph as pg

from .utilities.table_handler import TableHandler
from .utilities.matplotlibview import MatplotlibView
from . import MICRO, LAMBDA, ANGSTROMS, DELTA
from . import combine_image, bin_image, auto_image, manual_image, settings_image, more_infos_image
from . import stats_table_image, stats_plot_image


class Initialization:

    def __init__(self, parent=None):
        self.parent = parent

    def all(self):
        self.pyqtgraph_combine()
        self.pyqtgraph_bin()
        self.plot_widgets()
        self.statusbar()
        self.splitter()
        self.table()
        self.labels()
        self.tab()
        self.combobox()
        self.widgets()

    def statusbar(self):
        self.parent.eventProgress = QProgressBar(self.parent.ui.statusbar)
        self.parent.eventProgress.setMinimumSize(20, 14)
        self.parent.eventProgress.setMaximumSize(540, 100)
        self.parent.eventProgress.setVisible(False)
        self.parent.ui.statusbar.addPermanentWidget(self.parent.eventProgress)
        self.parent.setStyleSheet("QStatusBar{padding-left:8px;color:red;font-weight:bold;}")

    def splitter(self):
        self.parent.ui.combine_horizontal_splitter.setSizes([200, 500])
        self.parent.ui.bin_horizontal_splitter.setSizes([300, 800])
        self.parent.ui.bin_vertical_splitter.setSizes([500, 50])

    def combobox(self):
        list_of_options = ['mean', 'median', 'std', 'min', 'max']
        self.parent.ui.bin_stats_comboBox.addItems(list_of_options)

    def table(self):
        # combine table
        o_table = TableHandler(table_ui=self.parent.ui.combine_tableWidget)
        column_sizes = [50, 50, 500]
        o_table.set_column_sizes(column_sizes=column_sizes)

        # bin auto table
        o_table = TableHandler(table_ui=self.parent.ui.bin_auto_tableWidget)
        column_sizes = [40, 35, 60, 115, 115]
        o_table.set_column_sizes(column_sizes=column_sizes)
        column_names = ['use?',
                        'bin #',
                        'file #',
                        'tof range (' + MICRO + "s)",
                        LAMBDA + " range (" + ANGSTROMS + ")"]
        o_table.set_column_names(column_names=column_names)

        # bin manual table
        o_table = TableHandler(table_ui=self.parent.ui.bin_manual_tableWidget)
        column_sizes = [35, 80, 130, 130]
        o_table.set_column_sizes(column_sizes=column_sizes)
        column_names = column_names[1:]
        o_table.set_column_names(column_names=column_names)

        # statistics
        o_table = TableHandler(table_ui=self.parent.ui.statistics_tableWidget)
        column_names = ['bin #',
                        'file #',
                        'tof range (' + MICRO + "s)",
                        LAMBDA + " range (" + ANGSTROMS + ")",
                        'mean',
                        'median',
                        'std',
                        'min',
                        'max',
                        ]
        o_table.set_column_names(column_names=column_names)
        column_sizes = [35, 80, 130, 130, 130, 130, 130, 130, 130]
        o_table.set_column_sizes(column_sizes=column_sizes)

    def labels(self):
        # combine tab
        self.parent.ui.combine_detector_offset_units.setText(MICRO + "s")
        self.parent.ui.bin_tof_radioButton.setText("TOF (" + MICRO + "s)")
        self.parent.ui.bin_lambda_radioButton.setText(LAMBDA + " (" + ANGSTROMS + ")")
        # bin tab
        self.parent.ui.bin_auto_log_file_index_radioButton.setText(DELTA + "file_index/file_index")
        self.parent.ui.bin_auto_log_tof_radioButton.setText(DELTA + "tof")
        self.parent.ui.bin_auto_log_lambda_radioButton.setText(DELTA + LAMBDA + "/" + LAMBDA)

        self.parent.ui.auto_linear_file_index_radioButton.setText(DELTA + " file index")
        self.parent.ui.auto_linear_tof_radioButton.setText(DELTA + " tof")
        self.parent.ui.auto_linear_lambda_radioButton.setText(DELTA + LAMBDA)
        self.parent.ui.bin_auto_linear_tof_units_label.setText(MICRO + 's')
        self.parent.ui.bin_auto_linear_lambda_units_label.setText(ANGSTROMS)

    def tab(self):
        self.parent.ui.combine_bin_tabWidget.setTabIcon(0, QIcon(combine_image))
        self.parent.ui.combine_bin_tabWidget.setTabIcon(1, QIcon(bin_image))
        self.parent.ui.bin_tabWidget.setTabIcon(0, QIcon(auto_image))
        self.parent.ui.bin_tabWidget.setTabIcon(1, QIcon(manual_image))
        self.parent.ui.combine_bottom_tabWidget.setTabIcon(2, QIcon(settings_image))
        self.parent.ui.stats_tabWidget.setTabIcon(0, QIcon(stats_table_image))
        self.parent.ui.stats_tabWidget.setTabIcon(1, QIcon(stats_plot_image))
        self.parent.ui.combine_bin_tabWidget.setTabEnabled(1, False)

    def plot_widgets(self):
        graphics_view_layout = QVBoxLayout()
        statistics_plot = MatplotlibView(self.parent)
        graphics_view_layout.addWidget(statistics_plot)
        self.parent.ui.statistics_plot_widget.setLayout(graphics_view_layout)
        self.parent.statistics_plot = statistics_plot

    def pyqtgraph_bin(self):
        bin_view = pg.PlotWidget(title="")
        bin_view.plot()
        self.parent.bin_profile_view = bin_view
        layout = QVBoxLayout()
        layout.addWidget(bin_view)
        self.parent.ui.bin_widget.setLayout(layout)

    def pyqtgraph_combine(self):
        area = DockArea()
        self.parent.ui.area = area
        d1 = Dock("Image Preview", size=(200, 300))
        d2 = Dock("ROI profile", size=(200, 100))

        area.addDock(d1, 'top')
        area.addDock(d2, 'bottom')

        # preview - top widget
        image_view = pg.ImageView(view=pg.PlotItem())
        image_view.ui.roiBtn.hide()
        image_view.ui.menuBtn.hide()
        self.parent.combine_image_view = image_view
        image_view.scene.sigMouseMoved.connect(self.parent.mouse_moved_in_combine_image_preview)
        d1.addWidget(image_view)

        # plot and x-axis radio buttons - bottom widgets
        profile = pg.PlotWidget(title="")
        profile.plot()
        self.parent.combine_profile_view = profile
        bottom_layout = QVBoxLayout()
        bottom_layout.addWidget(profile)
        # xaxis radio buttons
        spacer_left = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        file_index_radio_button = QRadioButton("File Index")
        file_index_radio_button.setChecked(True)
        self.parent.combine_file_index_radio_button = file_index_radio_button
        self.parent.combine_file_index_radio_button.clicked.connect(self.parent.combine_xaxis_changed)
        tof_radio_button = QRadioButton("TOF (" + MICRO + "s)")
        self.parent.tof_radio_button = tof_radio_button
        self.parent.tof_radio_button.clicked.connect(self.parent.combine_xaxis_changed)
        lambda_radio_button = QRadioButton(LAMBDA + " (" + ANGSTROMS + ")")
        self.parent.lambda_radio_button = lambda_radio_button
        self.parent.lambda_radio_button.clicked.connect(self.parent.combine_xaxis_changed)
        spacer_right = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        axis_layout = QHBoxLayout()
        axis_layout.addItem(spacer_left)
        axis_layout.addWidget(file_index_radio_button)
        axis_layout.addWidget(tof_radio_button)
        axis_layout.addWidget(lambda_radio_button)
        axis_layout.addItem(spacer_right)
        bottom_widget = QWidget()
        bottom_widget.setLayout(axis_layout)
        bottom_layout.addWidget(bottom_widget)
        widget = QWidget()
        widget.setLayout(bottom_layout)
        d2.addWidget(widget)

        vertical_layout = QVBoxLayout()
        vertical_layout.addWidget(area)
        self.parent.ui.combine_widget.setLayout(vertical_layout)
        self.parent.ui.combine_widget.setEnabled(False)

    def widgets(self):
        self.parent.ui.visualize_auto_bins_axis_generated_pushButton.setIcon(QIcon(more_infos_image))
        self.parent.ui.visualize_auto_bins_axis_generated_pushButton.setToolTip("Display full original bin axis")
        self.parent.ui.bin_settings_pushButton.setIcon(QIcon(settings_image))
        self.parent.ui.combine_refresh_top_folder_pushButton.setEnabled(False)



TO_MICROS_UNITS = 1e6
TO_ANGSTROMS_UNITS = 1e10


class StatisticsName:

    mean = 'mean'
    median = 'median'
    std = 'std'
    min = 'min'
    max = 'max'


class StatisticsRegion:

    full = 'full'
    roi = 'roi'



import os
from qtpy.uic import loadUi
from maverick import ui
from ._version import get_versions

__version__ = get_versions()['version']
del get_versions

root = os.path.dirname(os.path.realpath(__file__))
refresh_image = os.path.join(root, "icons/refresh.png")
settings_image = os.path.join(root, "icons/plotSettings.png")
combine_image = os.path.join(root, "icons/combine.png")
bin_image = os.path.join(root, "icons/bin.png")
auto_image = os.path.join(root, "icons/auto.png")
manual_image = os.path.join(root, "icons/manual.png")
more_infos_image = os.path.join(root, "icons/more_infos.png")
stats_table_image = os.path.join(root, "icons/stats_table.png")
stats_plot_image = os.path.join(root, "icons/stats_plot.png")

ANGSTROMS = u"\u212B"
LAMBDA = u"\u03BB"
MICRO = u"\u00B5"
SUB_0 = u"\u2080"
DELTA = u"\u0394"


def load_ui(ui_filename, baseinstance):
    ui_filename = os.path.split(ui_filename)[-1]
    ui_path = os.path.dirname(ui.__file__)

    # get the location of the ui directory
    # this function assumes that all ui files are there
    filename = os.path.join(ui_path, ui_filename)

    return loadUi(filename, baseinstance=baseinstance)



from PIL import Image
import numpy as np
from tqdm import tqdm
import glob
import json
import os
import shutil
import copy
from pathlib import Path
from PIL.TiffTags import TAGS_V2 as TIFFTAGS_V2

from neutronbraggedge.experiment_handler.tof import TOF

from utilities.file_handler import FileHandler


class CombineBinCLI:

    algorithm = None

    def __init__(self, list_of_folders):
        folder_list_of_files_dict = CombineBinCLI.retrieve_list_of_files(list_of_folders)
        self.load_time_spectra_file(folder=list_of_folders[0])
        self.data_3d, self.metadata_array = CombineBinCLI.load_list_of_files(folder_list_of_files_dict)

    def combine(self, algorithm):
        self.algorithm = np.mean
        if algorithm == 'median':
            self.algorithm = np.median
        else:
            raise NotImplementedError(f"algorithm {algorithm} not implemented!")
        self.data_2d = self.algorithm(self.data_3d, axis=0)

    def load_time_spectra_file(self, folder=None):
        spectra_file = CombineBinCLI.get_spectra_file(folder)
        _tof_handler = TOF(filename=spectra_file)
        self.tof_array_s = _tof_handler.tof_array
        self.counts_array = _tof_handler.counts_array

    def bin(self, bin_table_file_name):
        with open(bin_table_file_name, 'r') as json_file:
            table = json.load(json_file)

        file_index_array = table['file_index_array']
        self.file_index_array = file_index_array
        nbr_file_index_array = len(file_index_array)
        list_full_image_rebinned = []
        for _index in tqdm(range(nbr_file_index_array)):
            _list_files = file_index_array[_index]
            nbr_files_for_that_bin = len(_list_files)
            data_to_work_with = []
            for _index_for_that_file in tqdm(range(nbr_files_for_that_bin)):
                _file_index = _list_files[_index_for_that_file]
                data_to_work_with.append(self.data_2d[_file_index])

            full_image_rebinned = self.algorithm(data_to_work_with, axis=0)
            list_full_image_rebinned.append(full_image_rebinned)

        self.data_2d_rebinned = list_full_image_rebinned

    def export(self, output_folder=None):
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)

        # data
        [nbr_files_to_create, height, width] = np.shape(self.data_2d_rebinned)
        counts_array = []
        for file_index in tqdm(range(nbr_files_to_create)):
            _output_file_name = str(Path(output_folder) / f"image_{file_index:04d}.tif")
            _data = self.data_2d_rebinned[file_index][:]
            counts_array.append(int(np.sum(_data)))
            _metadata = self.metadata_array[file_index]
            image = Image.fromarray(_data)
            image.save(_output_file_name, tiffinfo=_metadata)

        self.export_time_spectra(output_folder=output_folder,
                                 counts_array=counts_array)

    def export_time_spectra(self, output_folder=None, counts_array=None):
        new_file_name = str(Path(output_folder) / "image_Spectra.txt")
        tof_array = self.tof_array_s

        new_tof_array = []
        for _list_files_index in self.file_index_array:
            list_tof = [tof_array[_index] for _index in _list_files_index]
            new_tof_array.append(np.mean(list_tof))

        file_content = ["shutter_time,counts"]
        for _tof, _counts in zip(new_tof_array, counts_array):
            file_content.append(f"{_tof},{_counts}")

        FileHandler.make_ascii_file(data=file_content,
                                    output_file_name=new_file_name)


    @staticmethod
    def get_spectra_file(folder_name):
        time_spectra = glob.glob(folder_name + "/*_Spectra.txt")
        if time_spectra and os.path.exists(time_spectra[0]):
            return time_spectra[0]
        return ""

    @staticmethod
    def load_list_of_files(folder_list_of_files_dict):

        data_3d = []
        metadata_array = []

        list_of_keys = list(folder_list_of_files_dict.keys())

        for _index in tqdm(range(len(list_of_keys))):
            _folder = list_of_keys[_index]
            list_of_files = folder_list_of_files_dict[_folder]

            data_array = []

            for _file_index in tqdm(range(len(list_of_files))):
                _file = list_of_files[_file_index]
                metadata_data_dict = CombineBinCLI.get_tiff_data(_file)
                if _index == 0:
                    metadata_array.append(metadata_data_dict['metadata'])
                _data = copy.deepcopy(metadata_data_dict['data'])
                data_array.append(_data)

            data_3d.append(data_array)

        return data_3d, metadata_array

    @staticmethod
    def retrieve_list_of_files(list_of_folders):
        """
        create a dictionary of keys being the folder name, value being the list of files in this folder

        :param list_of_folders:
        :return: dictionary
        """
        if len(list_of_folders) < 2:
            raise ValueError("At least 2 folders are needed!")

        folder_list_of_files_dict = {}
        nbr_files_in_folder = -1
        for _folder_index in tqdm(range(len(list_of_folders))):
            _folder = list_of_folders[_folder_index]
            list_of_tif = FileHandler.get_list_of_tif(_folder)
            if nbr_files_in_folder == -1:
                nbr_files_in_folder = len(list_of_tif)
            else:
                if nbr_files_in_folder != len(list_of_tif):
                    raise ValueError("Folders do not have the same amount of files!")
                elif nbr_files_in_folder == 0:
                    raise ValueError(f"Folders {_folder} is empty!")

            folder_list_of_files_dict[_folder] = list_of_tif

        return folder_list_of_files_dict

    @staticmethod
    def get_tiff_data(filename):
        _o_image = Image.open(filename)

        # metadata
        metadata = CombineBinCLI.parse_tiff_header(_o_image)

        # image
        data = np.array(_o_image)
        _o_image.close()

        return {'metadata': metadata,
                'data': copy.deepcopy(data)}

    @staticmethod
    def parse_tiff_header(tiff_image):
        """Returns the tag from a TIFF image (loaded by PIL) in human readable dict."""
        return dict(tiff_image.tag_v2)



class TimeSpectraKeys:

    tof_array = 'tof_array'
    lambda_array = 'lambda_array'
    file_index_array = 'file_index_array'
    file_name = 'file_name'
    counts_array = 'counts_array'


class CombineAlgorithm:
    """list of algorithm used to combine the folders"""

    mean = 'mean'
    median = 'median'


class BinMode:
    """list of mode to bin the data"""

    auto = 'auto'
    manual = 'manual'
    settings = 'settings'


class BinAutoMode:

    log = 'log'
    linear = 'linear'


class BinAlgorithm:
    """list of algorithm used to bin the images"""

    mean = 'mean'
    median = 'median'



import logging

from ..session import SessionKeys
from ..utilities.get import Get
from ..utilities import TimeSpectraKeys, BinAutoMode, BinMode
from .manual_event_handler import ManualEventHandler
from .auto_event_handler import AutoEventHandler

FILE_INDEX_BIN_MARGIN = 0.5

UNSELECTED_BIN = (0, 0, 200, 50)
SELECTED_BIN = (0, 200, 0, 50)


class EventHandler:

    tof_bin_margin = 0
    lambda_bin_margin = 0

    def __init__(self, parent=None):
        self.parent = parent
        self.logger = logging.getLogger('maverick')

        self.tof_bin_margin = (self.parent.time_spectra[TimeSpectraKeys.tof_array][1] -
                               self.parent.time_spectra[TimeSpectraKeys.tof_array][0]) / 2.

        self.lambda_bin_margin = (self.parent.time_spectra[TimeSpectraKeys.lambda_array][1] -
                                  self.parent.time_spectra[TimeSpectraKeys.lambda_array][0]) / 2

    def entering_tab(self):
        o_get = Get(parent=self.parent)
        if o_get.bin_mode() == BinMode.auto:
            o_auto_event = AutoEventHandler(parent=self.parent)
            if o_get.bin_auto_mode() == BinAutoMode.linear:
                o_auto_event.auto_linear_radioButton_changed()
            elif o_get.bin_auto_mode() == BinAutoMode.log:
                o_auto_event.auto_log_radioButton_changed()
            o_auto_event.refresh_auto_tab()

        elif o_get.bin_mode() == BinMode.manual:
            o_manual_event = ManualEventHandler(parent=self.parent)
            o_manual_event.refresh_manual_tab()
            o_manual_event.display_all_items()

        else:
            pass

    def bin_auto_manual_tab_changed(self, new_tab_index=0):
        if new_tab_index == 0:
            self.parent.session[SessionKeys.bin_mode] = BinMode.auto

        elif new_tab_index == 1:
            self.parent.session[SessionKeys.bin_mode] = BinMode.manual

        elif new_tab_index == 2:
            pass

        else:
            raise NotImplementedError("LinearBin mode not implemented!")

        self.entering_tab()

    def bin_axis_changed(self):
        o_get = Get(parent=self.parent)
        if o_get.bin_mode() == BinMode.auto:
            o_event = AutoEventHandler(parent=self.parent)
            o_event.refresh_auto_tab()
        else:
            o_event = ManualEventHandler(parent=self.parent)
            o_event.refresh_manual_tab()



from qtpy.QtWidgets import QDialog
import os

from .. import load_ui
from ..session import SessionKeys, CombineAlgorithm


class Settings(QDialog):

    def __init__(self, parent=None):
        self.parent = parent
        QDialog.__init__(self, parent=parent)
        ui_full_path = os.path.join(os.path.dirname(__file__),
                                    os.path.join('ui', 'ui_bin_settings.ui'))
        self.ui = load_ui(ui_full_path, baseinstance=self)
        self.initialization()

    def initialization(self):
        if self.parent.session[SessionKeys.bin_algorithm] == CombineAlgorithm.mean:
            self.ui.bin_mean_radioButton.setChecked(True)
        elif self.parent.session[SessionKeys.bin_algorithm] == CombineAlgorithm.median:
            self.ui.bin_median_radioButton.setChecked(True)
        else:
            raise NotImplementedError("radio button bin algo not implemented!")

    def bin_algorithm_changed(self):
        if self.ui.bin_mean_radioButton.isChecked():
            self.parent.session[SessionKeys.bin_algorithm] = CombineAlgorithm.mean
        elif self.ui.bin_median_radioButton.isChecked():
            self.parent.session[SessionKeys.bin_algorithm] = CombineAlgorithm.median
        else:
            raise NotImplementedError("bin algorithm not implemented!")

        self.parent.update_statistics()



import numpy as np
import pyqtgraph as pg

from ..utilities import CombineAlgorithm
from ..session import SessionKeys
from ..utilities.get import Get


class Combine:

    def __init__(self, parent=None):
        self.parent = parent

    def run(self):
        combine_algorithm = self.parent.session[SessionKeys.combine_algorithm]
        # get list of data to combine
        o_get = Get(parent=self.parent)
        list_array_to_combine = o_get.list_array_to_combine()

        if list_array_to_combine == []:
            self.parent.combine_image_view.clear()
            self.parent.combine_image_view.removeItem(self.parent.combine_roi_item_id)
            self.parent.combine_data = None

        else:
            # combine using algorithm defined
            [nbr_folder_to_combine, nbr_files, width, height] = np.shape(list_array_to_combine)

            if nbr_folder_to_combine > 1:
                if combine_algorithm == CombineAlgorithm.mean:
                    combine_arrays = np.mean(list_array_to_combine, axis=0)
                elif combine_algorithm == CombineAlgorithm.median:
                    combine_arrays = np.median(list_array_to_combine, axis=0)
                else:
                    raise NotImplementedError("Algorithm not implemented!")
            else:
                combine_arrays = list_array_to_combine[0]

            combine_arrays = np.squeeze(combine_arrays)

            self.parent.combine_data = combine_arrays
            integrated_arrays = np.mean(combine_arrays, axis=0)
            integrated_arrays = np.transpose(integrated_arrays)
            self.parent.live_combine_image = integrated_arrays
            # display integrated
            self.parent.combine_image_view.setImage(integrated_arrays)

            # initialize ROI if first time, otherwise use same region
            [x0, y0, width, height] = self.parent.session[SessionKeys.combine_roi]
            if self.parent.combine_roi_item_id:
                self.parent.combine_image_view.removeItem(self.parent.combine_roi_item_id)

            roi_item = pg.ROI([x0, y0],
                              [width, height])
            roi_item.addScaleHandle([1, 1], [0, 0])
            self.parent.combine_image_view.addItem(roi_item)
            roi_item.sigRegionChanged.connect(self.parent.combine_roi_changed)
            self.parent.combine_roi_item_id = roi_item



import os
from os.path import expanduser
from pathlib import Path
import configparser
import tomli
import copy
import numpy as np

from ..utilities.table_handler import TableHandler
from . import CombineAlgorithm, TimeSpectraKeys, BinAutoMode, BinMode, BinAlgorithm
from ..session import SessionKeys
from ..bin import StatisticsName


class Get:

    def __init__(self, parent=None):
        self.parent = parent

    def log_file_name(self):
        log_file_name = self.parent.config['log_file_name']
        full_log_file_name = Get.full_home_file_name(log_file_name)
        return full_log_file_name

    def automatic_config_file_name(self):
        config_file_name = self.parent.config['session_file_name']
        full_config_file_name = Get.full_home_file_name(config_file_name)
        return full_config_file_name

    def combine_algorithm(self):
        if self.parent.ui.combine_mean_radioButton.isChecked():
            return CombineAlgorithm.mean
        elif self.parent.ui.combine_median_radioButton.isChecked():
            return CombineAlgorithm.median
        else:
            raise NotImplementedError("Combine algorithm not implemented!")

    def combine_x_axis_selected(self):
        if self.parent.combine_file_index_radio_button.isChecked():
            return TimeSpectraKeys.file_index_array
        elif self.parent.tof_radio_button.isChecked():
            return TimeSpectraKeys.tof_array
        elif self.parent.lambda_radio_button.isChecked():
            return TimeSpectraKeys.lambda_array
        else:
            raise NotImplementedError("xaxis not implemented in the combine tab!")

    def bin_mode(self):
        if self.parent.ui.bin_tabWidget.currentIndex() == 0:
            return BinMode.auto
        elif self.parent.ui.bin_tabWidget.currentIndex() == 1:
            return BinMode.manual
        elif self.parent.ui.bin_tabWidget.currentIndex() == 2:
            return BinMode.settings
        else:
            raise NotImplementedError("bin mode not implemented!")

    def bin_auto_mode(self):
        if self.parent.ui.auto_log_radioButton.isChecked():
            return BinAutoMode.log
        elif self.parent.ui.auto_linear_radioButton.isChecked():
            return BinAutoMode.linear
        else:
            raise NotImplementedError("auto bin mode not implemented!")

    def bin_x_axis_selected(self):
        if self.parent.ui.bin_file_index_radioButton.isChecked():
            return TimeSpectraKeys.file_index_array
        elif self.parent.ui.bin_tof_radioButton.isChecked():
            return TimeSpectraKeys.tof_array
        elif self.parent.ui.bin_lambda_radioButton.isChecked():
            return TimeSpectraKeys.lambda_array
        else:
            raise NotImplementedError("xaxis not implemented in bin tab!")

    def current_bins_activated(self):
        """
        Looking at the active auto or manual tab and linear or log to figure out which bins
        are currently used in the displayed
        :return: dictionary of the bins to use. Can be either self.parent.linear_bins, self.parent.log_bins
        or self.parent.manual_bins
        """
        bin_mode = self.bin_mode()
        if bin_mode == BinMode.manual:
            return self.parent.manual_bins
        elif bin_mode == BinMode.auto:
            bin_auto_mode = self.bin_auto_mode()
            if bin_auto_mode == BinAutoMode.linear:
                return self.parent.linear_bins
            elif bin_auto_mode == BinAutoMode.log:
                return self.parent.log_bins
            else:
                raise NotImplementedError("bin auto mode not implemented")
        else:
            raise NotImplementedError("bin mode not implemented!")

    def current_bins_name_activated(self):
        bin_mode = self.bin_mode()
        if bin_mode == BinMode.manual:
            return bin_mode
        else:
            return self.bin_auto_mode()

    def auto_log_bin_requested(self):
        if self.parent.ui.bin_auto_log_file_index_radioButton.isChecked():
            return self.parent.ui.auto_log_file_index_spinBox.value()
        elif self.parent.ui.bin_auto_log_tof_radioButton.isChecked():
            return self.parent.ui.auto_log_tof_doubleSpinBox.value()
        elif self.parent.ui.bin_auto_log_lambda_radioButton.isChecked():
            return self.parent.ui.auto_log_lambda_doubleSpinBox.value()
        else:
            raise NotImplementedError(f"auto log bin algorithm not implemented!")

    def auto_bins_currently_activated(self):
        auto_bin_mode = self.bin_auto_mode()
        if auto_bin_mode == BinAutoMode.linear:
            return self.parent.linear_bins
        elif auto_bin_mode == BinAutoMode.log:
            return self.parent.log_bins
        else:
            raise NotImplementedError("Auto bin mode not implemented!")

    def bin_log_axis(self):
        if self.parent.ui.bin_auto_log_file_index_radioButton.isChecked():
            return TimeSpectraKeys.file_index_array
        elif self.parent.ui.bin_auto_log_tof_radioButton.isChecked():
            return TimeSpectraKeys.tof_array
        elif self.parent.ui.bin_auto_log_lambda_radioButton.isChecked():
            return TimeSpectraKeys.lambda_array
        else:
            raise NotImplementedError(f"type not supported")

    def bin_linear_axis(self):
        if self.parent.ui.auto_linear_file_index_radioButton.isChecked():
            return TimeSpectraKeys.file_index_array
        elif self.parent.ui.auto_linear_tof_radioButton.isChecked():
            return TimeSpectraKeys.tof_array
        elif self.parent.ui.auto_linear_lambda_radioButton.isChecked():
            return TimeSpectraKeys.lambda_array
        else:
            raise NotImplementedError(f"type not supported")

    def bin_add_method(self):
        return self.parent.session[SessionKeys.bin_algorithm]

    def current_bin_tab_working_axis(self):
        bin_mode = self.bin_auto_mode()
        if bin_mode == BinAutoMode.log:
            return self.bin_log_axis()
        elif bin_mode == BinAutoMode.linear:
            return self.bin_linear_axis()
        else:
            raise NotImplementedError(f"type not supported")

    def bin_statistics_plot_requested(self):
        current_index = self.parent.ui.bin_stats_comboBox.currentIndex()
        list_name = [StatisticsName.mean,
                     StatisticsName.median,
                     StatisticsName.std,
                     StatisticsName.min,
                     StatisticsName.max]
        return list_name[current_index]

    def list_array_to_combine(self):
        session = self.parent.session
        list_working_folders_status = session[SessionKeys.list_working_folders_status]
        raw_data_folders = self.parent.raw_data_folders
        list_working_folders = session[SessionKeys.list_working_folders]

        if list_working_folders is None:
            return

        list_array = []
        for _status, _folder_name in zip(list_working_folders_status, list_working_folders):
            if _status:
                list_array.append(copy.deepcopy(raw_data_folders[_folder_name]['data']))

        return list_array

    def list_of_folders_to_use(self):
        o_table = TableHandler(table_ui=self.parent.ui.combine_tableWidget)
        nbr_row = o_table.row_count()
        list_of_folders_to_use = []
        list_of_folders_to_use_status = []
        for _row_index in np.arange(nbr_row):
            _horizontal_widget = o_table.get_widget(row=_row_index,
                                                    column=0)
            radio_button = _horizontal_widget.layout().itemAt(1).widget()
            if radio_button.isChecked():
                list_of_folders_to_use.append(o_table.get_item_str_from_cell(row=_row_index,
                                                                             column=2))
                status = True
            else:
                status = False
            list_of_folders_to_use_status.append(status)

        self.parent.session[SessionKeys.list_working_folders_status] = list_of_folders_to_use_status

        return list_of_folders_to_use

        # session = self.parent.session
        # list_working_folders_status = session[SessionKeys.list_working_folders_status]
        # list_working_folders = np.array(session[SessionKeys.list_working_folders])
        # try:
        #     return_list = list_working_folders[list_working_folders_status]
        # except IndexError:
        #     return [None]

    def manual_working_row(self, working_item_id=None):
        list_item_id = self.parent.list_of_manual_bins_item
        for _row, item in enumerate(list_item_id):
            if item == working_item_id:
                return _row
        return -1

    @staticmethod
    def full_home_file_name(base_file_name):
        home_folder = expanduser("~")
        full_log_file_name = os.path.join(home_folder, base_file_name)
        return full_log_file_name

    @staticmethod
    def version():
        setup_cfg = 'pyproject.toml'
        this_folder = os.path.abspath(os.path.dirname(__file__))
        top_path = Path(this_folder).parent.parent
        full_path_setup_cfg = str(Path(top_path) / Path(setup_cfg))

        ## to read from pyproject.toml file
        with open(full_path_setup_cfg, 'rb') as fp:
            config = tomli.load(fp)
        version = config['project']['version']

        ## to read from setup.cfg file
        # config = configparser.ConfigParser()
        # config.read(full_path_setup_cfg)
        # version = config['project']['version']
        return version



from qtpy.QtWidgets import QFileDialog
import logging
import json

from ..session import SessionKeys
from ..utilities.status_message_config import StatusMessageStatus, show_status_message
from ..bin.manual_event_handler import ManualEventHandler as BinManualEventHandler


class LoadBinTable:

    def __init__(self, parent=None):
        self.parent = parent
        self.logger = logging.getLogger("maverick")

    def run(self):
        working_dir = self.parent.session[SessionKeys.top_folder]
        bin_table = QFileDialog.getOpenFileName(caption="Select table bin",
                                                directory=working_dir,
                                                filter="table bin (*.json)")

        bin_table_file_name = bin_table[0]
        if not bin_table_file_name:
            self.logger.info("User cancel loading bin file!")
            return

        with open(bin_table_file_name, 'r') as json_file:
            table = json.load(json_file)

        show_status_message(parent=self.parent,
                            message=f"Bin table loaded from file {bin_table_file_name}",
                            status=StatusMessageStatus.ready,
                            duration_s=10)
        self.logger.info(f"Loaded {bin_table_file_name}!")

        self.parent.manual_bins = table

        o_event = BinManualEventHandler(parent=self.parent)
        o_event.clear_all_items()
        o_event.populate_table_with_this_table(table=table)
        self.parent.update_statistics()



import copy

from ..utilities.get import Get
from ..utilities import TimeSpectraKeys

from . import TO_MICROS_UNITS, TO_ANGSTROMS_UNITS
from .. import LAMBDA, MICRO, ANGSTROMS


class Plot:

    def __init__(self, parent=None):
        self.parent = parent

    def refresh_profile_plot(self):
        """
        this clear, remove all bin items and just replot the profile on the bin right imageView
        """

        self.parent.bin_profile_view.clear()  # clear previous plot
        if not (self.parent.dict_of_bins_item is None):  # remove previous bins
            for _key in self.parent.dict_of_bins_item.keys():
                self.parent.bin_profile_view.removeItem(self.parent.dict_of_bins_item[_key])

        profile_signal = self.parent.profile_signal

        o_get = Get(parent=self.parent)
        combine_algorithm = o_get.combine_algorithm()
        time_spectra_x_axis_name = o_get.bin_x_axis_selected()

        x_axis = copy.deepcopy(self.parent.time_spectra[time_spectra_x_axis_name])

        if time_spectra_x_axis_name == TimeSpectraKeys.file_index_array:
            x_axis_label = "file index"
        elif time_spectra_x_axis_name == TimeSpectraKeys.tof_array:
            x_axis *= TO_MICROS_UNITS    # to display axis in micros
            x_axis_label = "tof (" + MICRO + "s)"
        elif time_spectra_x_axis_name == TimeSpectraKeys.lambda_array:
            x_axis *= TO_ANGSTROMS_UNITS    # to display axis in Angstroms
            x_axis_label = LAMBDA + "(" + ANGSTROMS + ")"

        self.parent.bin_profile_view.plot(x_axis, profile_signal, pen='r', symbol='x')
        self.parent.bin_profile_view.setLabel("left", f"{combine_algorithm} counts")
        self.parent.bin_profile_view.setLabel("bottom", x_axis_label)



import subprocess


algorithm = 'median'
table_bin_file = '/Volumes/Buffalo/IPTS/IPTS-30023-Matteo-Simon/scan4/log_bin_table.json'
output_folder = '/Users/j35/Desktop/test_maverick_cli'
input_folders = ['/Volumes/Buffalo/IPTS/IPTS-30023-Matteo-Simon/scan4/Run_56306/',
                 '/Volumes/Buffalo/IPTS/IPTS-30023-Matteo-Simon/scan4/Run_56307/']
str_input_folders = " ".join(input_folders)

cmd = f"python -m maverick/maverick_combine_and_bin_cli.py -algorithm {algorithm} " \
      f"{table_bin_file} {output_folder} {str_input_folders}"
proc = subprocess.Popen(cmd,
                        shell=True,
                        stdin=subprocess.PIPE,
                        universal_newlines=True,
                        cwd=output_folder)
proc.communicate()



import argparse

from .combine_bin_cli import CombineBinCLI


parser = argparse.ArgumentParser(description='''
Combine and bin a set of TOF folders,

Example:
    python maverick/maverick_combine_cli.py -algorithm mean log_table_bin.json ./ /folder1 /folder2 /folder3 
    python maverick/maverick_combine_cli.py log_table_bin.json ./ /folder1 /folder2 /folder3
    python maverick/maverick_combine_cli.py log_table_bin.json ./ /folder3 /folder2
''',
                                 formatter_class=argparse.RawDescriptionHelpFormatter,
                                 epilog="NB: the list of input folders arguments must be last!")
parser.add_argument('-algorithm',
                    choices=["mean", "median"],
                    default="median",
                    help="Algorithm to use to combine the folders (default is median)")
parser.add_argument('bin_table_file_name',
                    help="file name of the table of bins created by the UI maverick")
parser.add_argument('export',
                    help="output folder where the combined data will be saved",
                    type=str)
parser.add_argument('folders',
                    help="list of folders to combine",
                    type=str,
                    nargs='+')
args = parser.parse_args()

# parsing arguments
algorithm = args.algorithm
bin_table_file_name = args.bin_table_file_name
export_folder = args.export
input_folders = args.folders

# load all the folders in memory
# combine using algorithm defined
# export all images into export folder using custom name
# move time stamp file from first folder to export folder using same custom name

# import time
# for i in tqdm(range(len(args.folders))):
#     time.sleep(1)

o_combine_bin = CombineBinCLI(list_of_folders=input_folders)
o_combine_bin.combine(algorithm)
o_combine_bin.bin(bin_table_file_name)
o_combine_bin.export(output_folder=export_folder)




## current command to run CLI and to test it using Buffalo
"""
python maverick/maverick_combine_and_bin_cli.py 
 /Volumes/Buffalo/IPTS/IPTS-30023-Matteo-Simon/scan4/log_bin_table.json 
 ~/Desktop/cli_output 
 /Volumes/Buffalo/IPTS/IPTS-30023-Matteo-Simon/scan4/Run_56306_light 
 /Volumes/Buffalo/IPTS/IPTS-30023-Matteo-Simon/scan4/Run_56307_light
"""


from ..log.log_launcher import LogHandler


class Check:

    def __init__(self, parent=None):
        self.parent = parent

    def log_file_size(self):
        o_handler = LogHandler(parent=self.parent,
                               log_file_name=self.parent.log_file_name)
        o_handler.cut_log_size_if_bigger_than_buffer()



import os
from qtpy.QtWidgets import QFileDialog, QCheckBox
import logging
import numpy as np
import copy

from ..utilities.file_handler import FileHandler
from ..utilities.table_handler import TableHandler
from ..utilities.time_spectra import GetTimeSpectraFilename, TimeSpectraHandler
from ..utilities.get import Get
from ..utilities import CombineAlgorithm, TimeSpectraKeys
from ..session import SessionKeys
from ..load.load_files import LoadFiles
from .combine import Combine
from .. import LAMBDA, MICRO, ANGSTROMS


class EventHandler:

    no_data_loaded = False

    def __init__(self, parent=None):
        self.parent = parent
        self.logger = logging.getLogger('maverick')

        o_get = Get(parent=parent)
        list_working_folders = o_get.list_of_folders_to_use()
        if list_working_folders == [None]:
            self.no_data_loaded = True

        # if not list_working_folders:
        #     self.no_data_loaded = True

        self.parent.session[SessionKeys.detector_offset] = self.parent.ui.detector_offset_spinBox.value()
        self.parent.session[SessionKeys.distance_source_detector] = \
            self.parent.ui.distance_source_detector_doubleSpinBox.value()
        self.parent.session[SessionKeys.sample_position] = self.parent.ui.combine_sample_position_doubleSpinBox.value()

    def select_top_folder(self):
        if self.no_data_loaded:
            return

        _folder = str(QFileDialog.getExistingDirectory(caption="Select Top Working Folder",
                                                       directory=self.parent.session[SessionKeys.top_folder],
                                                       options=QFileDialog.ShowDirsOnly))
        if _folder == "":
            self.logger.info("User Canceled the selection of top folder dialog!")
            return

        self.logger.info(f"Users selected a new top folder: {_folder}")

        # get list of folders in top folder
        self.parent.session[SessionKeys.top_folder] = os.path.abspath(_folder)
        list_folders = FileHandler.get_list_of_folders(_folder)
        self.parent.session[SessionKeys.list_working_folders] = list_folders

        # initialize parameters when using new working folder
        self.reset_data()

        # display the full path of the top folder selected
        self.parent.ui.top_folder_label.setText(_folder)

        # display list of folders in widget and column showing working folders used
        self.populate_list_of_folders_to_combine()

        # update ui
        self.check_widgets()

    def check_widgets(self):
        o_get = Get(parent=self.parent)

        if (o_get.list_of_folders_to_use() == []) or (o_get.list_of_folders_to_use() is None):
            self.parent.ui.combine_refresh_top_folder_pushButton.setEnabled(False)
            self.parent.ui.combine_bin_tabWidget.setTabEnabled(1, False)
            self.parent.ui.combine_widget.setEnabled(False)
        else:
            self.parent.ui.combine_refresh_top_folder_pushButton.setEnabled(True)
            self.parent.ui.combine_bin_tabWidget.setTabEnabled(1, True)
            self.parent.ui.combine_widget.setEnabled(True)

    def refresh_table_clicked(self):
        self.logger.info("User clicked the refresh table!")
        top_folder = self.parent.session[SessionKeys.top_folder]
        list_folders = FileHandler.get_list_of_folders(top_folder)
        # checking if there is any new folder
        current_list_of_folders = self.parent.session[SessionKeys.list_working_folders]
        for _folder in list_folders:
            if not (_folder in current_list_of_folders):
                self.parent.session[SessionKeys.list_working_folders].append(_folder)
                self.parent.session[SessionKeys.list_working_folders_status].append(False)

                list_files = FileHandler.get_list_of_files(_folder)
                nbr_files = len(list_files)
                data = {'data': None,
                        'list_files': list_files,
                        'nbr_files': nbr_files}
                self.parent.raw_data_folders[_folder] = data
                self.insert_row_entry(_folder)

    def reset_data(self):
        """
        This re-initialize all the parameters when working with a new top folder
        """

        # reset master dictionary that contains the raw data
        list_folders = self.parent.session[SessionKeys.list_working_folders]
        if list_folders is None:
            return

        _data_dict = {}
        for _folder in list_folders:
            list_files = FileHandler.get_list_of_tif(_folder)
            nbr_files = len(list_files)
            _data_dict[_folder] = {'data': None,
                                   'list_files': list_files,
                                   'nbr_files': nbr_files,
                                   }
        self.parent.raw_data_folders = _data_dict

        # initialize list of selected folders
        self.parent.session[SessionKeys.list_working_folders_status] = [False for _index in np.arange(len(list_folders))]

        # reset time spectra
        self.parent.time_spectra = {TimeSpectraKeys.file_name: None,
                                    TimeSpectraKeys.tof_array: None,
                                    TimeSpectraKeys.lambda_array: None,
                                    TimeSpectraKeys.file_index_array: None}

        self._reset_time_spectra_tab()

    def _reset_time_spectra_tab(self):
        self.parent.ui.time_spectra_name_label.setText("N/A")
        self.parent.ui.time_spectra_preview_pushButton.setEnabled(False)

    def populate_list_of_folders_to_combine(self):
        list_of_folders = self.parent.session[SessionKeys.list_working_folders]
        o_table = TableHandler(table_ui=self.parent.ui.combine_tableWidget)
        o_table.remove_all_rows()

        if list_of_folders is None:
            return

        for _folder in list_of_folders:
            self.insert_row_entry(folder=_folder)

    def insert_row_entry(self, folder=None):
        list_of_folders_status = self.parent.session.get(SessionKeys.list_working_folders_status, None)
        raw_data_folders = self.parent.raw_data_folders

        o_table = TableHandler(table_ui=self.parent.ui.combine_tableWidget)
        row = o_table.row_count()
        o_table.insert_empty_row(row=row)

        # use or not that row
        check_box = QCheckBox()
        if list_of_folders_status is None:
            status = False
        else:
            status = list_of_folders_status[row]
        check_box.setChecked(status)

        # check if this file has more than 1 file
        # if not disable radio button
        # number of images in that folder
        nbr_files = raw_data_folders[folder]['nbr_files']
        if nbr_files > 0:
            check_box.setEnabled(True)
        else:
            check_box.setEnabled(False)

        o_table.insert_widget(row=row,
                              column=0,
                              widget=check_box,
                              centered=True)
        check_box.clicked.connect(self.parent.radio_buttons_of_folder_changed)

        o_table.insert_item(row=row,
                            column=1,
                            value=nbr_files,
                            editable=False)

        # full path of the folder
        o_table.insert_item(row=row,
                            column=2,
                            value=folder,
                            editable=False)

    def update_list_of_folders_to_use(self, force_recalculation_of_time_spectra=False):
        o_get = Get(parent=self.parent)
        list_of_folders_to_use = o_get.list_of_folders_to_use()

        # o_table = TableHandler(table_ui=self.parent.ui.combine_tableWidget)
        # nbr_row = o_table.row_count()
        # list_of_folders_to_use = []
        # list_of_folders_to_use_status = []
        # for _row_index in np.arange(nbr_row):
        #     _horizontal_widget = o_table.get_widget(row=_row_index,
        #                                             column=0)
        #     radio_button = _horizontal_widget.layout().itemAt(1).widget()
        #     if radio_button.isChecked():
        #         list_of_folders_to_use.append(o_table.get_item_str_from_cell(row=_row_index,
        #                                                                      column=2))
        #         status = True
        #     else:
        #         status = False
        #     list_of_folders_to_use_status.append(status)
        #
        # self.parent.session[SessionKeys.list_working_folders_status] = list_of_folders_to_use_status

        self.logger.info("Updating list of folders to use:")
        self.logger.info(f"{list_of_folders_to_use}")

        # check or load the selected rows
        loading_worked = True

        for _folder_name in list_of_folders_to_use:

            if force_recalculation_of_time_spectra:
                self.load_time_spectra_file(folder=_folder_name)
                self.fix_linear_bin_radio_button_max_values()

            if self.parent.raw_data_folders[_folder_name]['data'] is None:
                loading_worked = self.load_that_folder(folder_name=_folder_name)

                # load time spectra if not already there
                if self.parent.time_spectra['file_name'] is None:
                    self.load_time_spectra_file(folder=_folder_name)
                    self.fix_linear_bin_radio_button_max_values()

    def load_time_spectra_file(self, folder=None):
        """
        load the time spectra file
        :param folder: location of the time spectra file
        :return:
        """
        o_time_spectra = GetTimeSpectraFilename(parent=self.parent,
                                                folder=folder)
        full_path_to_time_spectra = o_time_spectra.retrieve_file_name()

        o_time_handler = TimeSpectraHandler(parent=self.parent,
                                            time_spectra_file_name=full_path_to_time_spectra)
        o_time_handler.load()
        o_time_handler.calculate_lambda_scale()

        tof_array = o_time_handler.tof_array
        lambda_array = o_time_handler.lambda_array
        file_index_array = np.arange(len(tof_array))

        self.parent.time_spectra[TimeSpectraKeys.file_name] = full_path_to_time_spectra
        self.parent.time_spectra[TimeSpectraKeys.tof_array] = tof_array
        self.parent.time_spectra[TimeSpectraKeys.lambda_array] = lambda_array
        self.parent.time_spectra[TimeSpectraKeys.file_index_array] = file_index_array
        self.parent.time_spectra[TimeSpectraKeys.counts_array] = o_time_handler.counts_array

        # update time spectra tab
        self.parent.ui.time_spectra_name_label.setText(os.path.basename(full_path_to_time_spectra))
        self.parent.ui.time_spectra_preview_pushButton.setEnabled(True)

    def load_that_folder(self, folder_name=None):
        """
        this routine load all the images of the selected folder
        :param folder_name: full path of the folder containing the images to load
        :return: True if the loading worked, False otherwise
        """
        if not os.path.exists(folder_name):
            self.logger.info(f"Unable to load data from folder {folder_name}")
            return False

        # load the data
        o_load = LoadFiles(parent=self.parent,
                           folder=folder_name)
        data = o_load.retrieve_data()
        self.parent.raw_data_folders[folder_name]['data'] = data

        return True

    def combine_algorithm_changed(self):
        if self.no_data_loaded:
            return

        o_get = Get(parent=self.parent)

        combine_algorithm = o_get.combine_algorithm()
        self.parent.session[SessionKeys.combine_algorithm] = combine_algorithm
        self.logger.info(f"Algorithm to combine changed to: {combine_algorithm}")
        self.combine_folders()
        self.display_profile()

    def combine_folders(self):
        if self.no_data_loaded:
            return

        o_combine = Combine(parent=self.parent)
        o_combine.run()

    def combine_roi_changed(self):
        live_combine_image = self.parent.live_combine_image
        image_view = self.parent.combine_image_view
        roi_item = self.parent.combine_roi_item_id

        region = roi_item.getArraySlice(live_combine_image,
                                        image_view.imageItem)
        x0 = region[0][0].start
        x1 = region[0][0].stop - 1
        y0 = region[0][1].start
        y1 = region[0][1].stop - 1

        width = x1 - x0
        height = y1 - y0

        self.parent.session[SessionKeys.combine_roi] = [x0, y0, width, height]

    def display_profile(self):
        if self.no_data_loaded:
            return

        combine_data = self.parent.combine_data

        if combine_data is None:
            self.parent.combine_profile_view.clear()
            return

        [x0, y0, width, height] = self.parent.session[SessionKeys.combine_roi]

        o_get = Get(parent=self.parent)
        combine_algorithm = o_get.combine_algorithm()
        time_spectra_x_axis_name = o_get.combine_x_axis_selected()

        profile_signal = [np.mean(_data[y0:y0 + height, x0:x0 + width]) for _data in combine_data]
        # if combine_algorithm == CombineAlgorithm.mean:
        #     profile_signal = [np.mean(_data[y0:y0+height, x0:x0+width]) for _data in combine_data]
        # elif combine_algorithm == CombineAlgorithm.median:
        #     profile_signal = [np.median(_data[y0:y0+height, x0:x0+width]) for _data in combine_data]
        # else:
        #     raise NotImplementedError("Combine algorithm not implemented!")

        self.parent.profile_signal = profile_signal
        self.parent.combine_profile_view.clear()
        x_axis = copy.deepcopy(self.parent.time_spectra[time_spectra_x_axis_name])

        if time_spectra_x_axis_name == TimeSpectraKeys.file_index_array:
            x_axis_label = "file index"
        elif time_spectra_x_axis_name == TimeSpectraKeys.tof_array:
            x_axis *= 1e6    # to display axis in micros
            x_axis_label = "tof (" + MICRO + "s)"
        elif time_spectra_x_axis_name == TimeSpectraKeys.lambda_array:
            x_axis *= 1e10    # to display axis in Angstroms
            x_axis_label = LAMBDA + "(" + ANGSTROMS + ")"

        self.parent.combine_profile_view.plot(x_axis, profile_signal, pen='r', symbol='x')
        self.parent.combine_profile_view.setLabel("left", f"{combine_algorithm} counts")
        self.parent.combine_profile_view.setLabel("bottom", x_axis_label)

    def fix_linear_bin_radio_button_max_values(self):
        time_spectra = self.parent.time_spectra

        file_index_array = time_spectra[TimeSpectraKeys.file_index_array]
        max_file_index = int(len(file_index_array))

        tof_array = time_spectra[TimeSpectraKeys.tof_array]
        max_tof = float(tof_array[-1]) * 1e6
        min_tof = float(tof_array[0]) * 1e6

        lambda_array = time_spectra[TimeSpectraKeys.lambda_array]
        max_lambda = float(lambda_array[-1]) * 1e10
        min_lambda = float(lambda_array[0]) * 1e10

        self.parent.ui.auto_linear_file_index_spinBox.setMaximum(max_file_index)
        self.parent.ui.auto_linear_tof_doubleSpinBox.setMaximum(max_tof-min_tof)
        self.parent.ui.auto_linear_lambda_doubleSpinBox.setMaximum(max_lambda-min_lambda)

        self.logger.info(f"Max values of bin linear spin box has been fixed:")
        self.logger.info(f"-> file index: {max_file_index}")
        self.logger.info(f"-> tof: {max_tof}")
        self.logger.info(f"-> lambda: {max_lambda}")
